# coding: utf8

import numpy as np
from numpy.polynomial.chebyshev import poly2cheb  # cf rescale_largest_eig_val

import scipy.linalg as la

import matplotlib.pyplot as plt

from dppy.beta_ensemble_polynomial_potential_core import (polynomial_in_negative_log_conditional_a_coef as P_a_cond,
    polynomial_in_negative_log_conditional_b_coef as P_b_cond,
    sampler_exact_convex_quartic,
    sampler_mala,
    equilibrium_x2m,
    equilibrium_x2_x4,
    equilibrium_ClItKr10,
    check_random_state)


class BetaEnsemblePolynomialPotential(object):
    """docstring for BetaEnsemblePolynomialPotential"""
    def __init__(self, beta, potential_V, name=None):

        self.beta = beta
        self.V = potential_V

        if not (self.beta > 0):
            raise ValueError('beta = {} <= 0'.format(beta))

        if self.V.order > 7:

            str_ = ['Polynomial potentials V are allowed up to degree 6',
                    'Given\n',
                    ' '.join(['g_{}={}'.format(n, g_n)
                              for n, g_n in enumerate(self.V.coef[::-1])])]
            raise ValueError(' '.join(str_))

        if self.V[5]:
            raise ValueError('Potentials V = ... + g_5 x^5 + ... not supported, given g_5={}'.format(self.V[5]))

        if self.V.order % 2:
            raise ValueError('V has odd degree deg(V)={}'.format(self.V.order))

        if self.V[self.V.order] < 0:
            raise ValueError('V = g_{} x^{} + ... with g_{} < 0'.format(*[self.V.order] * 3))

        if self.V[0] != 0:
            print('Weird thing to introduce a constant V = ... + g_0={}. Reset g_0 = 0'.format(self.V[0]))
            self.V[0] = 0

        self._V_ClItKr10 = np.poly1d([1 / 20, -4 / 15, 1 / 5, 8 / 5, 0])

        self.equilibrium_density, self.support = self.__compute_equilibrium()

        # Sampling
        self.N = None
        self.nb_gibbs_passes = None

    def __str__(self):
        return '\n'.join(['beta={}'.format(self.beta),
                          'V(x) =\n{}'.format(self.V.__str__())])

    def __compute_equilibrium(self):
        """ Update attribute equilibrium_density and support if available
        """

        equi_dens, supp = None, None

        deg_V = self.V.order
        deg_V_2, deg_V_odd = divmod(deg_V, 2)

        set_non_zero_coefs = set(np.nonzero(self.V.coef[::-1])[0])

        if not deg_V_odd:
            if self.V == self._V_ClItKr10:
                equi_dens, supp = equilibrium_ClItKr10()
            elif set_non_zero_coefs == {deg_V}:
                equi_dens, supp = equilibrium_x2m(deg_V_2,
                                                  deg_V * self.V[deg_V])
            elif set_non_zero_coefs == {2, 4}:
                equi_dens, supp = equilibrium_x2_x4(2 * self.V[2],
                                                    4 * self.V[4])

        return equi_dens, supp

    def sample_mcmc(self,
                    N=10,
                    nb_gibbs_passes=10,
                    sample_exact_cond=False,
                    nb_mala_steps=100,
                    return_chain_of_eig_vals=False,
                    return_chain_of_lambda_max=False,
                    random_state=None):
        """ Gibbs sampler on Jacobi matrices to sample approximately from the corresponding :math:`\\beta`-ensemble.

        :param N:
            Number of points/size of the :math:`\\beta`-ensemble
        :type N:
            int

        :param nb_gibbs_passes:
            Number of passes/sweeps over the variables using the Gibbs sampler
        :type nb_gibbs_passes:
            int

        :param sample_exact_cond:
            Flag to force (``True``) exact sampling from the conditionals when it is possible.
            Otherwise run MALA for ``nb_mala_steps` to sample from the conditionals.
        :type sample_exact_cond:
            bool (default 100)

        :param nb_mala_steps:
            Number of steps of Metropolis Ajusted Langevin Algorithm (MALA) to perform when the conditionals are sampled approximately
        :type nb_mala_steps:
            int, default 100

        :param return_chain_of_eig_vals:
            Flag to return the chain of eigenvalues associated to the chain of Jacobi matrices.
            If ``True`` the whole chain of eigenvalues is returned
            If ``False`` only the last sequence of eigenvalues is returned
        :type return_chain_of_eig_vals:
            bool (default False)

        :param return_chain_of_lambda:
            Flag to return the chain of the **largest** eigenvalues associated to the chain of Jacobi matrices.
            If ``True`` the whole chain of the **largest** eigenvalues is returned
            If ``False`` only the **largest** eigenvalue of the last Jacobi matrix is returned
        :type return_chain_of_eig_vals:
            bool (default False)
        """

        rng = check_random_state(random_state)

        if sample_exact_cond:
            if self.V[3]:
                raise ValueError('Sampling exactly the conditionals a_i |... from V = ... + x^3 + ... is not supported, given g_3={}. Conditionals are not log-concave, cannot use Dev12 sampler'.format(self.V[3]))

            if self.V.order >= 5:
                raise ValueError('Sampling exactly the conditionals a_i |... from V = ... + x^5 + ... is not supported, deg(V)={}>=5. Conditionals are not log-concave, cannot use Dev12 sampler'.format(self.V.order))

            even_coefs_V = self.V.coef[::-1][2::2]
            if not all(even_coefs_V >= 0):
                raise ValueError('\n'.join(
                    ['even coefs of V are not all >=0',
                     ', '.join(['g_{}={}'.format(2 * (n + 1), g_2n)
                                for n, g_2n in enumerate(even_coefs_V)]),
                     'Conditionals are not log-concave, cannot use Dev12 sampler',
                     'You may retry swithching `sample_exact_cond` to False']))

        self.N = N
        self.nb_gibbs_passes = nb_gibbs_passes

        a, b = np.zeros((2, N + 3))

        if return_chain_of_eig_vals:
            eig_vals = np.zeros((N, nb_gibbs_passes))
        elif return_chain_of_lambda_max:
            lambda_max = np.zeros(nb_gibbs_passes)

        for p in range(nb_gibbs_passes):
            if (p + 1) % 50 == 0:
                print(p + 1)

            for i in range(1, N + 1):

                # a_i | ... propto exp - P_a_i
                P_a_i = 0.5 * self.beta * N * P_a_cond(i, a, b, self.V)
                if sample_exact_cond:
                    a[i], _ = sampler_exact_convex_quartic(
                                P=P_a_i,
                                random_state=rng)
                else:
                    a[i] = sampler_mala(a[i],
                                        V=P_a_i,
                                        sigma=0.01,
                                        nb_steps=nb_mala_steps,
                                        random_state=rng)

                # b_i | ... propto x^(shape-1) * exp - P_b_i
                if i < N:
                    P_b_i = 0.5 * self.beta * N * P_b_cond(i, a, b, self.V)
                    b[i], _ = sampler_exact_convex_quartic(
                                P=P_b_i,
                                shape=0.5 * self.beta * (N - i),
                                random_state=rng)

            if return_chain_of_eig_vals:
                eig_vals[:, p] = la.eigvalsh_tridiagonal(a[1:N + 1],
                                                         np.sqrt(b[1:N]))
            elif return_chain_of_lambda_max:
                lambda_max[p] = la.eigvalsh_tridiagonal(
                                    a[1:N + 1],
                                    np.sqrt(b[1:N]),
                                    select='i',
                                    select_range=(N - 1, N - 1))[0]

        if return_chain_of_eig_vals:
            return eig_vals
        if return_chain_of_lambda_max:
            return lambda_max

        return la.eigvalsh_tridiagonal(a[1:N + 1], np.sqrt(b[1:N]))

    def hist(self, sampl, save_file_name=False):
        """ Display the histogram of a ``sampl`` from the corresponding :math:`\\beta`-ensemble and the corresponding equilibrium distribution when available

        :param sampl:
            One or multiple samples from the corresponding :math:`\\beta`-ensemble.
            **In any case ``sampl`` is flattened** as if the samples were concatenated
        :type sampl:
            array_like

        :param save_file_name:
            File name, e.g. ``figure.pdf``, to save the plot
        :type save_file_name:
            str

        .. seealso::

            :py:func:`__compute_equilibrium`
        """

        fig, ax = plt.subplots(1, 1)

        # Title
        # V_x = ' '.join(['V(x) =',
        #                 ' + '.join([r'$g_{} x^{}$'.format(n, n)
        #                             for n, g_n in enumerate(self.V.coef[::-1],
        #                                                     start=0)
        #                             if g_n])])

        # with_coefs = ', '.join([r'$g_{}={:0.2f}$'.format(n, g_n)
        #                         for n, g_n in enumerate(self.V.coef[::-1])
        #                         if g_n])

        # beta_N_passes = r'$\beta={}, N={}$ #Gibbs_passes={}'.format(
        #                 self.beta, self.N, self.nb_gibbs_passes)
        # plt.title('\n'.join([V_x, with_coefs, beta_N_passes]))

        # histogram
        ax.hist(np.ravel(sampl),
                density=True,
                histtype='step',
                lw=3,
                bins=30,
                label='histogram')

        if self.equilibrium_density is not None and self.support is not None:
            # equilibrium_measure
            x = np.linspace(1.1 * self.support[0],
                            1.1 * self.support[1],
                            300)
            ax.plot(x, self.equilibrium_density(x),
                    label=r'$\mu_{eq}$', lw=3, c='k')

        # start, end = ax.get_xlim()
        # ax.xaxis.set_ticks(np.arange(-1.5, 2.1, 1.5))
        ax.xaxis.set_ticks(np.arange(-2, 2.1, 1))

        plt.legend(loc='best',
                   fontsize='x-large',
                   frameon=False,
                   handlelength=1)
        ax.tick_params(axis='both', which='major', labelsize=15)
        plt.tight_layout()

        if save_file_name:
            plt.savefig(save_file_name)

    def rescale_largest_eig_val(self, lambda_max):
        """ Rescale the largest eigenvalue to see Tracy-Widom fluctuations

        .. math::

            N^{\\frac{2}{3}} c_v (\\lambda_{\\max} - b_v)

        where

        .. math::

            c_V = (b_v - a_v)^{-\\frac{1}{3}} \\left(\\sum_{n=1}^{\\infty} k {V'}_k \\right)^{\\frac{2}{3}}

        with :math:`{V'}_k` being the Chebychev coefficients of

        .. math::

            V'(\\frac{a_v + b_v}{2} + \\frac{b_v - a_v}{2} X)

        .. seealso::

            - Section 3.2 https://arxiv.org/pdf/1210.2199.pdf
            - :cite:`OlNaTr14` p.5 Equation 2.3 `https://arxiv.org/pdf/1404.0071.pdf <https://arxiv.org/pdf/1404.0071.pdf>`_
            - `poly2cheb <https://docs.scipy.org/doc/numpy/reference/generated/numpy.polynomial.chebyshev.poly2cheb.html?>`_
        """

        a_v, b_v = self.support
        shift = np.poly1d([0.5 * (b_v - a_v), 0.5 * (b_v + a_v)])

        dV_shift = self.V.deriv(m=1)(shift)

        dV_shift_cheb = poly2cheb(dV_shift.coeffs[::-1])

        sum_k_dV_k = sum(k * dV_k
                         for k, dV_k in enumerate(dV_shift_cheb[1:], start=1))

        c_v = np.cbrt(sum_k_dV_k**2 / (b_v - a_v))

        return self.N**(2 / 3) * c_v * (lambda_max - b_v)



# coding: utf8
""" Implementation of :py:class:`FiniteDPP` object which has 6 main methods:

- :py:meth:`~FiniteDPP.sample_exact`, see also :ref:`sampling DPPs exactly<finite_dpps_exact_sampling>`
- :py:meth:`~FiniteDPP.sample_exact_k_dpp`, see also :ref:`sampling k-DPPs exactly<finite_dpps_exact_sampling>`
- :py:meth:`~FiniteDPP.sample_mcmc`, see also :ref:`sampling DPPs with MCMC<finite_dpps_mcmc_sampling>`
- :py:meth:`~FiniteDPP.sample_mcmc_k_dpp`, see also :ref:`sampling k-DPPs with MCMC<finite_dpps_mcmc_sampling_k_dpps>`
- :py:meth:`~FiniteDPP.compute_K`, to compute the correlation :math:`K` kernel from initial parametrization
- :py:meth:`~FiniteDPP.compute_L`, to compute the likelihood :math:`L` kernel from initial parametrization

.. seealso:

    `Documentation on ReadTheDocs <https://dppy.readthedocs.io/en/latest/finite_dpps/index.html>`_
"""

import numpy as np
import scipy.linalg as la
import matplotlib.pyplot as plt

from warnings import warn

from dppy.exact_sampling import (dpp_sampler_generic_kernel,
                                 proj_dpp_sampler_kernel,
                                 proj_dpp_sampler_eig,
                                 dpp_vfx_sampler,
                                 dpp_eig_vecs_selector,
                                 k_dpp_vfx_sampler,
                                 k_dpp_eig_vecs_selector,
                                 elementary_symmetric_polynomials)

from dppy.mcmc_sampling import (dpp_sampler_mcmc,
                                zonotope_sampler)

from dppy.utils import (check_random_state,
                        is_symmetric,
                        is_projection,
                        is_orthonormal_columns,
                        is_full_row_rank,
                        is_in_01,
                        is_geq_0,
                        is_equal_to_O_or_1)


class FiniteDPP:
    """ Finite DPP object parametrized by

    :param string kernel_type:

        - ``'correlation'`` :math:`\\mathbf{K}` kernel
        - ``'likelihood'`` :math:`\\mathbf{L}` kernel

    :param projection:
        Indicate whether the provided kernel is of projection type. This may be useful when the
        :class:`FiniteDPP` object is defined through its correlation kernel :math:`\\mathbf{K}`.

    :type projection:
        bool, default ``False``

    :param params:
        Dictionary containing the parametrization of the underlying

        - correlation kernel

            - ``{'K': K}``, with :math:`0 \\preceq \\mathbf{K} \\preceq I`
            - ``{'K_eig_dec': (eig_vals, eig_vecs)}``, with :math:`0 \\leq eigvals \\leq 1`
            - ``{'A_zono': A}``, with :math:`A (d \\times N)` and :math:`\\operatorname{rank}(A)=d`

        - likelihood kernel

            - ``{'L': L}``, with :math:`\\mathbf{L}\\succeq 0`
            - ``{'L_eig_dec': (eig_vals, eig_vecs)}``, with :math:`eigvals \\geq 0`
            - ``{'L_gram_factor': Phi}``, with :math:`\\mathbf{L} = \\Phi^{ \\top} \\Phi`
            - ``{'L_eval_X_data': (eval_L, X_data)}``, with :math:`\\mathbf{X}_{data}(N \\times d)` and
              :math:`eval \\_ L` a likelihood function such that
              :math:`\\mathbf{L} = eval \\_ L(\\mathbf{X}_{data}, \\{X}_{data})`. For a full description of the
              requirements imposed on `eval_L`'s interface, see the documentation :func:`dppy.vfx_sampling.vfx_sampling_precompute_constants`.
              For an example, see the implementation of any of the kernels provided by scikit-learn
              (e.g. sklearn.gaussian_process.kernels.PairwiseKernel).



    :type params:
        dict

    .. caution::

        For now we only consider real valued matrices :math:`\\mathbf{K}, \\mathbf{L}, A, \\Phi, \\mathbf{X}_{data}`.

    .. seealso::

        - :ref:`finite_dpps_definition`
        - :ref:`finite_dpps_exact_sampling`

    .. todo::

        add ``.kernel_rank`` attribute
    """

    ###############
    # Constructor #
    ###############
    def __init__(self, kernel_type, projection=False, **params):

        self.kernel_type = kernel_type
        self.projection = projection
        self.params_keys = set(params.keys())
        self.__check_args_coherence()

        # Sampling
        self.sampling_mode = 'GS'  # Gram-Schmidt
        self.list_of_samples = []

        # when using .sample_k_dpp_*
        self.size_k_dpp = 0
        self.E_poly = None  # evaluation of the

        # Attributes relative to K correlation kernel:
        # K, K_eig_vals, K_eig_vecs, A_zono
        self.K = is_symmetric(params.get('K', None))
        if self.projection:
            self.K = is_projection(self.K)

        e_vals, e_vecs = params.get('K_eig_dec', [None, None])
        if self.projection:
            self.K_eig_vals = is_equal_to_O_or_1(e_vals)
        else:
            self.K_eig_vals = is_in_01(e_vals)
        self.eig_vecs = is_orthonormal_columns(e_vecs)

        self.A_zono = is_full_row_rank(params.get('A_zono', None))

        # Attributes relative to L likelihood kernel:
        # L, L_eig_vals, L_eig_vecs, L_gram_factor, L_dual
        self.L = is_symmetric(params.get('L', None))
        if self.projection:
            self.L = is_projection(self.L)

        e_vals, e_vecs = params.get('L_eig_dec', [None, None])
        if self.projection:
            self.L_eig_vals = is_equal_to_O_or_1(e_vals)
        else:
            self.L_eig_vals = is_geq_0(e_vals)
        if self.eig_vecs is None:  # K_eig_vecs = L_eig_vecs
            self.eig_vecs = is_orthonormal_columns(e_vecs)

        # L' "dual" likelihood kernel, L' = Phi Phi.T, Phi = L_gram_factor
        self.L_gram_factor = params.get('L_gram_factor', None)
        self.L_dual = None

        if self.L_gram_factor is not None:
            Phi = self.L_gram_factor
            d, N = Phi.shape
            if d < N:
                self.L_dual = Phi.dot(Phi.T)
                print('L_dual = Phi Phi.T was computed: Phi (dxN) with d<N')
            else:
                if self.L is None:
                    self.L = Phi.T.dot(Phi)
                    print('L = Phi.T Phi was computed: Phi (dxN) with d>=N')

        # L likelihood function representation
        # eval_L(X, Y) = L(X, Y)
        # eval_L(X) = L(X, X)
        self.eval_L, self.X_data = params.get('L_eval_X_data', [None, None])
        self.intermediate_sample_info = None

        if self.eval_L is not None:
            if not callable(self.eval_L):
                raise ValueError('eval_L should be a positive semi-definite kernel function')
        if self.X_data is not None:
            if not (self.X_data.size and self.X_data.ndim == 2):
                err_print = ['Wrong shape = {}'.format(self.X_data.shape),
                             'X_data should be a non empty (N x d) ndarray']
                raise ValueError('\n'.join(err_print))

    def __str__(self):
        str_info = ['DPP defined through {} {} kernel'
                    .format('projection' if self.projection else '',
                            self.kernel_type),
                    'Parametrized by {}'
                    .format(self.params_keys),
                    '- sampling mode = {}'
                    .format(self.sampling_mode),
                    '- number of samples = {}'
                    .format(len(self.list_of_samples))]

        return '\n'.join(str_info)

    # Check routine
    def __check_args_coherence(self):
        # Check coherence of initialization parameters of the DPP:
        # kernel_type, projection and params.

        if not isinstance(self.projection, bool):
            err_print = ['`projection` argument is not boolean',
                         'Given: {}'.format(self.is_projection)]
            raise ValueError('\n'.join(err_print))

        K_type, K_params = 'correlation', {'K', 'K_eig_dec', 'A_zono'}
        L_type, L_params = 'likelihood', {'L', 'L_eig_dec', 'L_gram_factor', 'L_eval_X_data'}

        if self.kernel_type == K_type:
            if self.params_keys.intersection(K_params):
                if 'A_zono' in self.params_keys and not self.projection:
                    warn_print = ['Weird setting:',
                                  'FiniteDPP(kernel_type={}, projection={}, **{"A_zono": A}) with projection=False',
                                  'When defined through "A_zono" we expect a projection DPP with correlation kernel K = A.T (AA.T)^-1 A`.',
                                  'projection` switched to `True`']
                    warn('\n'.join(warn_print))
                    self.projection = True
            else:
                err_print =\
                    ['Invalid parametrization of correlation kernel, choose:',
                     '- {"K": K} 0 <= K <= I',
                     '- {"K_eig_dec": (e_vals, e_vecs)} 0 <= e_vals <= 1',
                     '- {"A_zono": A} A (dxN) s.t. K = A.T (AA.T)^-1 A',
                     'Given: {}'.format(self.params_keys)]
                raise ValueError('\n'.join(err_print))

        elif self.kernel_type == L_type:
            if self.params_keys.intersection(L_params):
                if self.projection:
                    warn('Weird setting: defining a DPP via a projection likelihood L kernel is unusual. Make sure you do not want to use a projection CORRELATION K kernel instead')
            else:
                err_print =\
                    ['Invalid parametrization of likelihood kernel, choose:',
                     '- {"L": L} L >= 0',
                     '- {"L_eig_dec": (e_vals, e_vecs)} e_vals >= 0',
                     '- {"L_gram_factor": Phi}, Phi (dxN) s.t. L = Phi.TPhi',
                     '- {"L_eval_X_data": (eval_L, X_data)} X_data (dxN) and `eval_L` callable positive semi-definite kernel',
                     'Given: {}'.format(self.params_keys)]
                raise ValueError('\n'.join(err_print))

        else:
            err_print =\
                ['Invalid `kernel_type`, choose among:',
                 '- `correlation`: K kernel',
                 '- `likelihood`: L kernel'
                 'Given: {}'.format(self.params_keys)]
            raise ValueError('\n'.join(err_print))

    ##################
    # Object methods #
    ##################
    def info(self):
        """ Display infos about the :class:`FiniteDPP` object
        """
        print(self.__str__())

    def flush_samples(self):
        """ Empty the :py:attr:`~FiniteDPP.list_of_samples` attribute.

        .. see also::

            - :py:meth:`~FiniteDPP.sample_exact`
            - :py:meth:`~FiniteDPP.sample_mcmc`
        """
        self.list_of_samples = []
        self.size_k_dpp = 0

    # Exact sampling
    def sample_exact(self, mode='GS', **params):
        """ Sample exactly from the corresponding :class:`FiniteDPP <FiniteDPP>` object.

        :param mode:

            - ``projection=True``:
                - ``'GS'`` (default): Gram-Schmidt on the rows of :math:`\\mathbf{K}`.
                - ``'Chol'`` :cite:`Pou19` Algorithm 3
                - ``'Schur'``: when DPP defined from correlation kernel ``K``, use Schur complement to compute conditionals

            - ``projection=False``:
                - ``'GS'`` (default): Gram-Schmidt on the rows of the eigenvectors of :math:`\\mathbf{K}` selected in Phase 1.
                - ``'GS_bis'``: Slight modification of ``'GS'``
                - ``'Chol'`` :cite:`Pou19` Algorithm 1
                - ``'KuTa12'``: Algorithm 1 in :cite:`KuTa12`
                - ``'vfx'``: the dpp-vfx rejection sampler in :cite:`DeCaVa19`

        :type mode:
            string, default ``'GS'``

        :param dict params:
            Dictionary containing the parameters for exact samplers with keys

            - ``'random_state'`` (default None)
            - If ``mode='vfx'``

                See :py:meth:`~dppy.exact_sampling.dpp_vfx_sampler` for a full list of all parameters accepted by 'vfx' sampling. We report here the most impactful

                + ``'rls_oversample_dppvfx'`` (default 4.0) Oversampling parameter used to construct dppvfx's internal Nystrom approximation. This makes each rejection round slower and more memory intensive, but reduces variance and the number of rounds of rejections.
                + ``'rls_oversample_bless'`` (default 4.0) Oversampling parameter used during bless's internal Nystrom approximation. This makes the one-time pre-processing slower and more memory intensive, but reduces variance and the number of rounds of rejections

                Empirically, a small factor [2,10] seems to work for both parameters. It is suggested to start with a small number and increase if the algorithm fails to terminate.

        :return:
            Returns a sample from the corresponding :class:`FiniteDPP <FiniteDPP>` object. In any case, the sample is appended to the :py:attr:`~FiniteDPP.list_of_samples` attribute as a list.

        :rtype:
            list

        .. note::

            Each time you call this method, the sample is appended to the :py:attr:`~FiniteDPP.list_of_samples` attribute as a list.

            The :py:attr:`~FiniteDPP.list_of_samples` attribute can be emptied using :py:meth:`~FiniteDPP.flush_samples`

        .. caution::

            The underlying kernel :math:`\\mathbf{K}`, resp. :math:`\\mathbf{L}` must be real valued for now.

        .. seealso::

            - :ref:`finite_dpps_exact_sampling`
            - :py:meth:`~FiniteDPP.flush_samples`
            - :py:meth:`~FiniteDPP.sample_mcmc`
        """

        rng = check_random_state(params.get('random_state', None))

        self.sampling_mode = mode

        if self.sampling_mode == 'Schur':
            if self.kernel_type == 'correlation' and self.projection:
                self.compute_K()
                sampl = proj_dpp_sampler_kernel(self.K, self.sampling_mode,
                                                random_state=rng)
            else:
                err_print =\
                    ['`Schur` sampling mode is only available for projection DPPs, i.e., `kernel_type="correlation"` and `projection=True`',
                     'Given: {}'.format((self.kernel_type, self.projection))]
                raise ValueError('\n'.join(err_print))

        elif self.sampling_mode == 'Chol':
            self.compute_K()
            if self.kernel_type == 'correlation' and self.projection:
                sampl = proj_dpp_sampler_kernel(self.K, self.sampling_mode,
                                                random_state=rng)
            else:
                sampl, _ = dpp_sampler_generic_kernel(self.K, random_state=rng)

        elif self.sampling_mode == 'vfx':
            if self.eval_L is None or self.X_data is None:
                raise ValueError('The vfx sampler is currently only available with '
                                 '{"L_eval_X_data": (L_eval, X_data)} representation.')

            params.pop("random_state", None)
            sampl, self.intermediate_sample_info = dpp_vfx_sampler(
                                                self.intermediate_sample_info,
                                                self.X_data,
                                                self.eval_L,
                                                random_state=rng,
                                                **params)

        # If eigen decoposition of K, L or L_dual is available USE IT!
        elif self.K_eig_vals is not None:
            # Phase 1
            if self.kernel_type == 'correlation' and self.projection:
                V = self.eig_vecs[:, self.K_eig_vals > 0.5]
            else:
                V = dpp_eig_vecs_selector(self.K_eig_vals, self.eig_vecs,
                                          random_state=rng)
            # Phase 2
            sampl = proj_dpp_sampler_eig(V, self.sampling_mode,
                                         random_state=rng)

        # elif self.L_dual_eig_vals is not None:
        #     # Phase 1
        #     V = dpp_eig_vecs_selector_L_dual(self.L_dual_eig_vals,
        #                                      self.L_dual_eig_vecs,
        #                                      self.L_gram_factor,
        #                                      random_state=rng)
        #     # Phase 2
        #     sampl = proj_dpp_sampler_eig(V, self.sampling_mode,
        #                                  random_state=rng)
        #

        elif self.L_eig_vals is not None:
            self.K_eig_vals = self.L_eig_vals / (1.0 + self.L_eig_vals)
            return self.sample_exact(mode=self.sampling_mode,
                                     random_state=rng)

        elif self.L_dual is not None:
            # L_dual = Phi Phi.T = W Theta W.T
            # L = Phi.T Phi = V Gamma V
            # implies Gamma = Theta and V = Phi.T W Theta^{-1/2}
            self.L_eig_vals, L_dual_eig_vecs = la.eigh(self.L_dual)
            self.L_eig_vals = is_geq_0(self.L_eig_vals)
            self.eig_vecs = self.L_gram_factor.T.dot(L_dual_eig_vecs
                                                    / np.sqrt(self.L_eig_vals))
            return self.sample_exact(mode=self.sampling_mode,
                                     random_state=rng)

        # If DPP defined via projection correlation kernel K
        # no eigendecomposition required
        elif self.K is not None and self.projection:
            sampl = proj_dpp_sampler_kernel(self.K, self.sampling_mode,
                                            random_state=rng)

        elif self.K is not None:
            self.K_eig_vals, self.eig_vecs = la.eigh(self.K)
            self.K_eig_vals = is_in_01(self.K_eig_vals)
            return self.sample_exact(mode=self.sampling_mode,
                                     random_state=rng)

        elif self.L is not None:
            self.L_eig_vals, self.eig_vecs = la.eigh(self.L)
            self.L_eig_vals = is_geq_0(self.L_eig_vals)
            return self.sample_exact(mode=self.sampling_mode,
                                     random_state=rng)

        # If DPP defined through correlation kernel with parameter 'A_zono'
        # a priori you wish to use the zonotope approximate sampler
        elif self.A_zono is not None:
            warn('DPP defined via `A_zono`, apriori you want to use `sample_mcmc`, but you have called `sample_exact`')

            self.K_eig_vals = np.ones(self.A_zono.shape[0])
            self.eig_vecs, _ = la.qr(self.A_zono.T, mode='economic')
            return self.sample_exact(mode=self.sampling_mode,
                                     random_state=rng)

        elif self.eval_L is not None and self.X_data is not None:
            self.compute_L()
            return self.sample_exact(mode=self.sampling_mode,
                                     random_state=rng)

        else:
            raise ValueError('None of the available samplers could be used based on the current DPP representation. This should never happen, please consider rasing an issue on github at https://github.com/guilgautier/DPPy/issues')

        self.list_of_samples.append(sampl)
        return sampl

    def sample_exact_k_dpp(self, size, mode='GS', **params):
        """ Sample exactly from :math:`\\operatorname{k-DPP}`. A priori the :class:`FiniteDPP <FiniteDPP>` object was instanciated by its likelihood :math:`\\mathbf{L}` kernel so that

        .. math::

            \\mathbb{P}_{\\operatorname{k-DPP}}(\\mathcal{X} = S)
                \\propto \\det \\mathbf{L}_S ~ 1_{|S|=k}

        :param size:
            size :math:`k` of the :math:`\\operatorname{k-DPP}`

        :type size:
            int

        :param mode:
            - ``projection=True``:
                - ``'GS'`` (default): Gram-Schmidt on the rows of :math:`\\mathbf{K}`.
                - ``'Schur'``: Use Schur complement to compute conditionals.

            - ``projection=False``:
                - ``'GS'`` (default): Gram-Schmidt on the rows of the eigenvectors of :math:`\\mathbf{K}` selected in Phase 1.
                - ``'GS_bis'``: Slight modification of ``'GS'``
                - ``'KuTa12'``: Algorithm 1 in :cite:`KuTa12`
                - ``'vfx'``: the dpp-vfx rejection sampler in :cite:`DeCaVa19`

        :type mode:
            string, default ``'GS'``

        :param dict params:
            Dictionary containing the parameters for exact samplers with keys

            ``'random_state'`` (default None)

            - If ``mode='vfx'``

                See :py:meth:`~dppy.exact_sampling.k_dpp_vfx_sampler` for a full list of all parameters accepted by 'vfx' sampling. We report here the most impactful

                + ``'rls_oversample_dppvfx'`` (default 4.0) Oversampling parameter used to construct dppvfx's internal Nystrom approximation. This makes each rejection round slower and more memory intensive, but reduces variance and the number of rounds of rejections.
                + ``'rls_oversample_bless'`` (default 4.0) Oversampling parameter used during bless's internal Nystrom approximation. This makes the one-time pre-processing slower and more memory intensive, but reduces variance and the number of rounds of rejections

                Empirically, a small factor [2,10] seems to work for both parameters. It is suggested to start with
                a small number and increase if the algorithm fails to terminate.

        :return:
            A sample from the corresponding :math:`\\operatorname{k-DPP}`.

            In any case, the sample is appended to the :py:attr:`~FiniteDPP.list_of_samples` attribute as a list.

        :rtype:
            list

        .. note::

            Each time you call this method, the sample is appended to the :py:attr:`~FiniteDPP.list_of_samples` attribute as a list.

            The :py:attr:`~FiniteDPP.list_of_samples` attribute can be emptied using :py:meth:`~FiniteDPP.flush_samples`

        .. caution::

            The underlying kernel :math:`\\mathbf{K}`, resp. :math:`\\mathbf{L}` must be real valued for now.

        .. seealso::

            - :py:meth:`~FiniteDPP.sample_exact`
            - :py:meth:`~FiniteDPP.sample_mcmc_k_dpp`
        """

        rng = check_random_state(params.get('random_state', None))

        self.sampling_mode = mode
        self.size_k_dpp = size

        if self.sampling_mode == 'vfx':
            if self.eval_L is None or self.X_data is None:
                raise ValueError("The vfx sampler is currently only available for the 'L_eval_X_data' representation.")

            params.pop("random_state", None)
            sampl, self.intermediate_sample_info = k_dpp_vfx_sampler(
                                                size,
                                                self.intermediate_sample_info,
                                                self.X_data,
                                                self.eval_L,
                                                random_state=rng,
                                                **params)

        # If DPP defined via projection kernel
        elif self.projection:
            if self.kernel_type == 'correlation':

                if self.K_eig_vals is not None:
                    rank = np.rint(np.sum(self.K_eig_vals)).astype(int)
                elif self.A_zono is not None:
                    rank = self.A_zono.shape[0]
                else:  # self.K is not None
                    rank = np.rint(np.trace(self.K)).astype(int)

                if size != rank:
                    raise ValueError('size k={} != rank={} for projection correlation K kernel'.format(size, rank))

                if self.K_eig_vals is not None:
                    # K_eig_vals > 0.5 below to get indices where e_vals = 1
                    sampl = proj_dpp_sampler_eig(
                            eig_vecs=self.eig_vecs[:, self.K_eig_vals > 0.5],
                            mode=self.sampling_mode,
                            size=size,
                            random_state=rng)

                elif self.A_zono is not None:
                    warn('DPP defined via `A_zono`, apriori you want to use `sampl_mcmc`, but you have called `sample_exact`')

                    self.K_eig_vals = np.ones(rank)
                    self.eig_vecs, _ = la.qr(self.A_zono.T, mode='economic')

                    sampl = proj_dpp_sampler_eig(eig_vecs=self.eig_vecs,
                                                 mode=self.sampling_mode,
                                                 size=size,
                                                 random_state=rng)

                else:
                    sampl = proj_dpp_sampler_kernel(kernel=self.K,
                                                    mode=self.sampling_mode,
                                                    size=size,
                                                    random_state=rng)

            else:  # self.kernel_type == 'likelihood':
                if self.L_eig_vals is not None:
                    # L_eig_vals > 0.5 below to get indices where e_vals = 1
                    sampl = proj_dpp_sampler_eig(
                            eig_vecs=self.eig_vecs[:, self.L_eig_vals > 0.5],
                            mode=self.sampling_mode,
                            size=size,
                            random_state=rng)
                else:
                    self.compute_L()
                    # size > rank treated internally in proj_dpp_sampler_kernel
                    sampl = proj_dpp_sampler_kernel(self.L,
                                                    mode=self.sampling_mode,
                                                    size=size,
                                                    random_state=rng)

        # If eigen decoposition of K, L or L_dual is available USE IT!
        elif self.L_eig_vals is not None:

            # Phase 1
            # Precompute elementary symmetric polynomials
            if self.E_poly is None or self.size_k_dpp < size:
                self.E_poly = elementary_symmetric_polynomials(self.L_eig_vals,
                                                               size)
            # Select eigenvectors
            V = k_dpp_eig_vecs_selector(self.L_eig_vals, self.eig_vecs,
                                        size=size,
                                        E_poly=self.E_poly,
                                        random_state=rng)
            # Phase 2
            self.size_k_dpp = size
            sampl = proj_dpp_sampler_eig(V, self.sampling_mode,
                                         random_state=rng)

        elif self.K_eig_vals is not None:
            np.seterr(divide='raise')
            self.L_eig_vals = self.K_eig_vals / (1.0 - self.K_eig_vals)
            return self.sample_exact_k_dpp(size, self.sampling_mode,
                                           random_state=rng)

        # Otherwise eigendecomposition is necessary
        elif self.L_dual is not None:
            # L_dual = Phi Phi.T = W Theta W.T
            # L = Phi.T Phi = V Gamma V.T
            # implies Gamma = Theta and V = Phi.T W Theta^{-1/2}
            self.L_eig_vals, L_dual_eig_vecs = la.eigh(self.L_dual)
            self.L_eig_vals = is_geq_0(self.L_eig_vals)
            self.eig_vecs =self.L_gram_factor.T.dot(L_dual_eig_vecs
                                                    / np.sqrt(self.L_eig_vals))
            return self.sample_exact_k_dpp(size, mode=self.sampling_mode,
                                           random_state=rng)

        elif self.L is not None:
            self.L_eig_vals, self.eig_vecs = la.eigh(self.L)
            self.L_eig_vals = is_geq_0(self.L_eig_vals)
            return self.sample_exact_k_dpp(size, self.sampling_mode,
                                           random_state=rng)

        elif self.K is not None:
            self.K_eig_vals, self.eig_vecs = la.eigh(self.K)
            self.K_eig_vals = is_in_01(self.K_eig_vals)
            return self.sample_exact_k_dpp(size, self.sampling_mode,
                                           random_state=rng)

        elif self.eval_L is not None and self.X_data is not None:
            # In case mode!='vfx'
            self.compute_L()
            return self.sample_exact_k_dpp(size, self.sampling_mode,
                                           random_state=rng)

        else:
            raise ValueError('None of the available samplers could be used based on the current DPP representation. This should never happen, please consider rasing an issue on github at https://github.com/guilgautier/DPPy/issues')

        self.list_of_samples.append(sampl)
        return sampl

    # Approximate sampling
    def sample_mcmc(self, mode, **params):
        """ Run a MCMC with stationary distribution the corresponding :class:`FiniteDPP <FiniteDPP>` object.

        :param string mode:

            - ``'AED'`` Add-Exchange-Delete
            - ``'AD'`` Add-Delete
            - ``'E'`` Exchange
            - ``'zonotope'`` Zonotope sampling

        :param dict params:
            Dictionary containing the parameters for MCMC samplers with keys

            ``'random_state'`` (default None)

            - If ``mode='AED','AD','E'``

                + ``'s_init'`` (default None) Starting state of the Markov chain
                + ``'nb_iter'`` (default 10) Number of iterations of the chain
                + ``'T_max'`` (default None) Time horizon
                + ``'size'`` (default None) Size of the initial sample for ``mode='AD'/'E'``

                    * :math:`\\operatorname{rank}(\\mathbf{K})=\\operatorname{trace}(\\mathbf{K})` for projection :math:`\\mathbf{K}` (correlation) kernel and ``mode='E'``

            - If ``mode='zonotope'``:

                + ``'lin_obj'`` linear objective in main optimization problem (default np.random.randn(N))
                + ``'x_0'`` initial point in zonotope (default A*u, u~U[0,1]^n)
                + ``'nb_iter'`` (default 10) Number of iterations of the chain
                + ``'T_max'`` (default None) Time horizon

        :return:
            The last sample of the trajectory of Markov chain.

            In any case, the full trajectory of the Markov chain, made of ``params['nb_iter']`` samples, is appended to the :py:attr:`~FiniteDPP.list_of_samples` attribute as a list of lists.

        :rtype:
            list

        .. note::

            Each time you call this method, the full trajectory of the Markov chain, made of ``params['nb_iter']`` samples, is appended to the :py:attr:`~FiniteDPP.list_of_samples` attribute as a list of lists.

            The :py:attr:`~FiniteDPP.list_of_samples` attribute can be emptied using :py:meth:`~FiniteDPP.flush_samples`

        .. seealso::

            - :ref:`finite_dpps_mcmc_sampling`
            - :py:meth:`~FiniteDPP.sample_exact`
            - :py:meth:`~FiniteDPP.flush_samples`
        """

        self.sampling_mode = mode

        if self.sampling_mode == 'zonotope':
            if self.A_zono is not None:
                chain = zonotope_sampler(self.A_zono, **params)
            else:
                err_print = ['Invalid `mode=zonotope` parameter',
                             'DPP must be defined via `A_zono`',
                             'Given: {}'.format(self.params_keys)]
                raise ValueError(' '.join(err_print))

        elif self.sampling_mode == 'E':
            if self.kernel_type == 'correlation' and self.projection:
                self.compute_K()
                size = params.get('size', None)
                rank = np.rint(np.trace(self.K)).astype(int)
                # |sample| = Tr(K) = rank(K) a.s. for projection DPP(K)
                if size == rank:
                    chain = dpp_sampler_mcmc(self.K,
                                             self.sampling_mode,
                                             **params)
                else:
                    raise ValueError('size={} != rank={} for projection correlation K kernel'.format(size, rank))
            else:
                self.compute_L()
                chain = dpp_sampler_mcmc(self.L, self.sampling_mode, **params)

        elif self.sampling_mode in ('AED', 'AD'):
            self.compute_L()
            chain = dpp_sampler_mcmc(self.L, self.sampling_mode, **params)

        else:
            err_print = ['Invalid `mode` parameter, choose among:',
                         '- `AED`: Add-Exchange-Delete',
                         '- `AD`: Add-Delete',
                         '- `E`: Exchange',
                         '- `zonotope`: projection correlation kernel only',
                         'Given: {}'.format(self.sampling_mode)]
            raise ValueError('\n'.join(err_print))

        self.list_of_samples.append(chain)
        return chain[-1]

    def sample_mcmc_k_dpp(self, size, mode='E', **params):
        """ Calls :py:meth:`~sample_mcmc` with ``mode='E'`` and ``params['size'] = size``

        .. seealso::

            - :ref:`finite_dpps_mcmc_sampling`
            - :py:meth:`~FiniteDPP.sample_mcmc`
            - :py:meth:`~FiniteDPP.sample_exact_k_dpp`
            - :py:meth:`~FiniteDPP.flush_samples`
        """

        self.sampling_mode = 'E'

        self.size_k_dpp = size
        params['size'] = size

        return self.sample_mcmc(self.sampling_mode, **params)

    def compute_K(self, msg=False):
        """ Compute the correlation kernel :math:`\\mathbf{K}` from the original parametrization of the :class:`FiniteDPP` object.

        The kernel is stored in the :py:attr:`~FiniteDPP.K` attribute.

        .. seealso::

            :ref:`finite_dpps_relation_kernels`
        """

        if self.K is not None:
            # msg = 'K (correlation) kernel available'
            # print(msg)
            pass

        else:
            if not msg:
                print('K (correlation) kernel computed via:')

            if self.K_eig_vals is not None:
                msg = '- U diag(eig_K) U.T'
                print(msg)
                self.K = (self.eig_vecs * self.K_eig_vals).dot(self.eig_vecs.T)

            elif self.A_zono is not None:
                msg = '\n'.join(['- K = A.T (AA.T)^-1 A, using',
                                 '- U = QR(A.T)',
                                 '- K = U U.T'])
                print(msg)
                self.K_eig_vals = np.ones(self.A_zono.shape[0])
                self.eig_vecs, _ = la.qr(self.A_zono.T, mode='economic')
                self.K = self.eig_vecs.dot(self.eig_vecs.T)

            elif self.L_eig_vals is not None:
                msg = '- eig_K = eig_L/(1+eig_L)'
                print(msg)
                self.K_eig_vals = self.L_eig_vals / (1.0 + self.L_eig_vals)
                self.compute_K(msg=True)

            elif self.L is not None:
                msg = '- eigendecomposition of L'
                print(msg)
                self.L_eig_vals, self.eig_vecs = la.eigh(self.L)
                self.L_eig_vals = is_geq_0(self.L_eig_vals)
                self.compute_K(msg=True)

            else:
                self.compute_L(msg=True)
                self.compute_K(msg=True)

    def compute_L(self, msg=False):
        """ Compute the likelihood kernel :math:`\\mathbf{L}` from the original parametrization of the :class:`FiniteDPP` object.

        The kernel is stored in the :py:attr:`~FiniteDPP.L` attribute.

        .. seealso::

            :ref:`finite_dpps_relation_kernels`
        """

        if self.L is not None:
            # msg = 'L (likelihood) kernel available'
            # print(msg)
            pass

        elif (self.kernel_type == 'correlation') and self.projection:
            err_print = ['L = K(I-K)^-1 = kernel cannot be computed:',
                         'K is projection kernel: some eigenvalues equal 1']
            raise ValueError('\n'.join(err_print))

        else:
            if not msg:
                print('L (likelihood) kernel computed via:')

            if self.L_eig_vals is not None:
                msg = '- U diag(eig_L) U.T'
                print(msg)
                self.L = (self.eig_vecs * self.L_eig_vals).dot(self.eig_vecs.T)

            elif self.L_gram_factor is not None:
                msg = '- L = Phi.T Phi, where Phi = L_gram_factor'
                print(msg)
                self.L = self.L_gram_factor.T.dot(self.L_gram_factor)

            elif self.eval_L is not None:
                warn_print = ['Weird setting:',
                'FiniteDPP(.., **{"L_eval_X_data": (eval_L, X_data)})',
                'When using "L_eval_X_data", you are a priori working with a big `X_data` and not willing to compute the full likelihood kernel L',
                'Right now, the computation of L=eval_L(X_data) is performed but might be very expensive, this is at your own risk!',
                'You might also use FiniteDPP(.., **{"L": eval_L(X_data)})']
                warn('\n'.join(warn_print))
                msg = '- L = eval_L(X_data, X_data)'
                print(msg)
                self.L = self.eval_L(self.X_data)

            elif self.K_eig_vals is not None:
                try:  # to compute eigenvalues of kernel L = K(I-K)^-1
                    msg = '- eig_L = eig_K/(1-eig_K)'
                    print(msg)
                    np.seterr(divide='raise')
                    self.L_eig_vals = self.K_eig_vals / (1.0 - self.K_eig_vals)
                    self.compute_L(msg=True)
                except FloatingPointError:
                    err_print = ['Eigenvalues of L kernel cannot be computed',
                                 'eig_L = eig_K/(1-eig_K)',
                                 'K kernel has some eig_K very close to 1',
                                 'Hint: `K` kernel might be a projection']
                    raise FloatingPointError('\n'.join(err_print))

            elif self.K is not None:
                msg = '- eigendecomposition of K'
                print(msg)
                self.K_eig_vals, self.eig_vecs = la.eigh(self.K)
                self.K_eig_vals = is_in_01(self.K_eig_vals)
                self.compute_L(msg=True)

            else:
                self.compute_K(msg=True)
                self.compute_L(msg=True)

    def plot_kernel(self, kernel_type='correlation', save_path=''):
        """Display a heatmap of the kernel used to define the :class:`FiniteDPP` object (correlation kernel :math:`\\mathbf{K}` or likelihood kernel :math:`\\mathbf{L}`)

        :param title:
            Plot title

        :type title:
            string
        """

        if not kernel_type:
            kernel_type = self.kernel_type

        fig, ax = plt.subplots(1, 1)

        if kernel_type == 'correlation':
            self.compute_K()
            nb_items, kernel_to_plot = self.K.shape[0], self.K
            str_title = r'$K$ (correlation) kernel'

        elif kernel_type == 'likelihood':
            self.compute_L()
            nb_items, kernel_to_plot = self.L.shape[0], self.L
            str_title = r'$L$ (likelihood) kernel'

        else:
            raise ValueError('kernel_type != "correlation" or "likelihood"')

        heatmap = ax.pcolor(kernel_to_plot, cmap='jet', vmin=-0.3, vmax=1)

        ax.set_aspect('equal')

        ticks = np.arange(nb_items)
        ticks_label = [r'${}$'.format(tic) for tic in ticks]

        ax.xaxis.tick_top()
        ax.set_xticks(ticks + 0.5, minor=False)

        ax.invert_yaxis()
        ax.set_yticks(ticks + 0.5, minor=False)

        ax.set_xticklabels(ticks_label, minor=False)
        ax.set_yticklabels(ticks_label, minor=False)

        # plt.title(title if title else str_title, y=1.1)

        plt.colorbar(heatmap)

        if save_path:
            plt.savefig(save_path)



# coding: utf8
""" Implementation of full and banded matrix models for :math:`\beta`-Ensemble:

- Hermite Ensemble (full + tridiagonal)
- Laguerre Ensemble (full + tridiagonal)
- Jacobi Ensemble (full + tridiagonal)
- Circular Ensemble (full + quindiagonal)
- Ginibre Ensemble (full)

.. seealso:

    `Documentation on ReadTheDocs <https://dppy.readthedocs.io/en/latest/continuous_dpps/beta_ensembles.sampling.html>`_
"""

import numpy as np
import scipy.linalg as la
import scipy.sparse as sp

from dppy.utils import check_random_state


###########
# Hermite #
###########
def hermite_sampler_full(N, beta=2,
                         random_state=None):

    rng = check_random_state(random_state)

    if beta == 1:
        A = rng.randn(N, N)

    elif beta == 2:
        A = rng.randn(N, N) + 1j * rng.randn(N, N)

    elif beta == 4:
        X = rng.randn(N, N) + 1j * rng.randn(N, N)
        Y = rng.randn(N, N) + 1j * rng.randn(N, N)

        A = np.block([[X, Y], [-Y.conj(), X.conj()]])

    else:
        err_print = ('`beta` parameter must be 1, 2 or 4.',
                     'Given: {}'.format(beta))
        raise ValueError('\n'.join(err_print))

    return la.eigvalsh(A + A.conj().T) / np.sqrt(2.0)


def semi_circle_law(x, R=2.0):
    """
    .. seealso::

        - :cite:`DuEd15` Table 1

        - https://en.wikipedia.org/wiki/Wigner_semicircle_distribution
    """
    return 2 / (np.pi * R**2) * np.sqrt(np.maximum(R**2 - x**2, 0.0))


def mu_ref_normal_sampler_tridiag(loc=0.0, scale=1.0, beta=2, size=10,
                                  random_state=None):
    """Implementation of the tridiagonal model to sample from

    .. math::

        \\Delta(x_{1}, \\dots, x_{N})^{\\beta}
        \\prod_{n=1}^{N} \\exp(-\\frac{(x_i-\\mu)^2}{2\\sigma^2} ) dx_i

    .. seealso::

        :cite:`DuEd02` II-C
    """

    rng = check_random_state(random_state)

    if not (beta > 0):
        raise ValueError('`beta` must be positive. Given: {}'.format(beta))

    # beta/2*[N-1, N-2, ..., 1]
    b_2_Ni = 0.5 * beta * np.arange(size - 1, 0, step=-1)

    alpha_coef = rng.normal(loc=loc, scale=scale, size=size)
    beta_coef = rng.gamma(shape=b_2_Ni, scale=scale**2)

    return la.eigvalsh_tridiagonal(alpha_coef, np.sqrt(beta_coef))


############
# Laguerre #
############
def laguerre_sampler_full(M, N, beta=2,
                          random_state=None):

    rng = check_random_state(random_state)

    if beta == 1:
        A = rng.randn(N, M)

    elif beta == 2:
        A = rng.randn(N, M) + 1j * rng.randn(N, M)

    elif beta == 4:
        X = rng.randn(N, M) + 1j * rng.randn(N, M)
        Y = rng.randn(N, M) + 1j * rng.randn(N, M)
        A = np.block([[X, Y], [-Y.conj(), X.conj()]])

    else:
        err_print = ('`beta` parameter must be 1, 2 or 4.',
                     'Given: {}'.format(beta))
        raise ValueError('\n'.join(err_print))

    return la.eigvalsh(A.dot(A.conj().T))


def marcenko_pastur_law(x, M, N, sigma=1.0):
    """ M >= N

    .. seealso::

        - :cite:`DuEd15` Table 1
        - https://en.wikipedia.org/wiki/Marchenko-Pastur_distribution
    """

    c = N / M
    Lm, Lp = (sigma * (1 - np.sqrt(c)))**2, (sigma * (1 + np.sqrt(c)))**2

    return np.sqrt(np.maximum((Lp-x)*(x-Lm),0)) / (c*x) / (2*np.pi*sigma**2)


def mu_ref_gamma_sampler_tridiag(shape=1.0, scale=1.0, beta=2, size=10,
                                 random_state=None):
    """
    .. seealso::

        :cite:`DuEd02` III-B
    """

    rng = check_random_state(random_state)

    if not (beta > 0):
        raise ValueError('`beta` must be positive. Given: {}'.format(beta))

    # beta/2*[N-1, N-2, ..., 1, 0]
    b_2_Ni = 0.5 * beta * np.arange(size - 1, -1, step=-1)

    # xi_odd = xi_1, ... , xi_2N-1
    xi_odd = rng.gamma(shape=b_2_Ni + shape, scale=scale)  # odd

    # xi_even = xi_0=0, xi_2, ... ,xi_2N-2
    xi_even = np.zeros(size)
    xi_even[1:] = rng.gamma(shape=b_2_Ni[:-1], scale=scale)  # even

    # alpha_i = xi_2i-2 + xi_2i-1, xi_0 = 0
    alpha_coef = xi_even + xi_odd
    # beta_i+1 = xi_2i-1 * xi_2i
    beta_coef = xi_odd[:-1] * xi_even[1:]

    return la.eigvalsh_tridiagonal(alpha_coef, np.sqrt(beta_coef))


##########
# Jacobi #
##########
def jacobi_sampler_full(M_1, M_2, N, beta=2,
                        random_state=None):

    rng = check_random_state(random_state)

    if beta == 1:
        X = rng.randn(N, M_1)
        Y = rng.randn(N, M_2)

    elif beta == 2:
        X = rng.randn(N, M_1) + 1j * rng.randn(N, M_1)
        Y = rng.randn(N, M_2) + 1j * rng.randn(N, M_2)

    elif beta == 4:
        X_1 = rng.randn(N, M_1) + 1j * rng.randn(N, M_1)
        X_2 = rng.randn(N, M_1) + 1j * rng.randn(N, M_1)

        Y_1 = rng.randn(N, M_2) + 1j * rng.randn(N, M_2)
        Y_2 = rng.randn(N, M_2) + 1j * rng.randn(N, M_2)

        X = np.block([[X_1, X_2], [-X_2.conj(), X_1.conj()]])
        Y = np.block([[Y_1, Y_2], [-Y_2.conj(), Y_1.conj()]])

    else:
        err_print = ('`beta` parameter must be 1, 2 or 4.',
                     'Given: {}'.format(beta))
        raise ValueError('\n'.join(err_print))

    X_tmp = X.dot(X.conj().T)
    Y_tmp = Y.dot(Y.conj().T)

    return la.eigvals(X_tmp.dot(la.inv(X_tmp + Y_tmp))).real


def wachter_law(x, M_1, M_2, N):
    """ M_1, M_2>=N

    .. seealso::

        :cite:`DuEd15` Table 1
    """
    a, b = M_1 / N, M_2 / N

    Lm = ((np.sqrt(a * (a + b - 1)) - np.sqrt(b)) / (a + b))**2
    Lp = ((np.sqrt(a * (a + b - 1)) + np.sqrt(b)) / (a + b))**2

    return (a+b)/(2*np.pi) * 1/(x*(1-x)) * np.sqrt(np.maximum((Lp-x)*(x-Lm),0))


def mu_ref_beta_sampler_tridiag(a, b, beta=2, size=10,
                                random_state=None):
    """ Implementation of the tridiagonal model given by Theorem 2 of :cite:`KiNe04` to sample from

    .. math::

        \\Delta(x_{1}, \\dots, x_{N})^{\\beta}
        \\prod_{n=1}^{N} x^{a-1} (1-x)^{b-1} dx

    .. seealso::

        :cite:`KiNe04` Theorem 2
    """

    rng = check_random_state(random_state)

    if not (beta > 0):
        raise ValueError('`beta` must be positive. Given: {}'.format(beta))

    # beta/2*[N-1, N-2, ..., 1, 0]
    b_2_Ni = 0.5 * beta * np.arange(size - 1, -1, step=-1)

    # c_odd = c_1, c_3, ..., c_2N-1
    c_odd = rng.beta(b_2_Ni + a, b_2_Ni + b)

    # c_even = c_0, c_2, c_2N-2
    c_even = np.zeros(size)
    c_even[1:] = rng.beta(b_2_Ni[:-1], b_2_Ni[1:] + a + b)

    # xi_odd = xi_2i-1 = (1-c_2i-2) c_2i-1
    xi_odd = (1 - c_even) * c_odd

    # xi_even = xi_0=0, xi_2, xi_2N-2
    # xi_2i = (1-c_2i-1)*c_2i
    xi_even = np.zeros(size)
    xi_even[1:] = (1 - c_odd[:-1]) * c_even[1:]

    # alpha_i = xi_2i-2 + xi_2i-1
    # alpha_1 = xi_0 + xi_1 = xi_1
    alpha_coef = xi_even + xi_odd
    # beta_i+1 = xi_2i-1 * xi_2i
    beta_coef = xi_odd[:-1] * xi_even[1:]

    return la.eigvalsh_tridiagonal(alpha_coef, np.sqrt(beta_coef))


#####################
# Circular ensemble #
#####################
def circular_sampler_full(N, beta=2, haar_mode='QR',
                          random_state=None):
    """
    .. seealso::

        :cite:`Mez06` Section 5
    """

    rng = check_random_state(random_state)

    if haar_mode == 'Hermite':
        # size_sym_mat = int(N*(N-1)/2)

        if beta == 1:  # COE
            A = rng.randn(N, N)

        elif beta == 2:  # CUE
            A = rng.randn(N, N) + 1j * rng.randn(N, N)

        elif beta == 4:
            X = rng.randn(N, N) + 1j * rng.randn(N, N)
            Y = rng.randn(N, N) + 1j * rng.randn(N, N)

            A = np.block([[X, Y], [-Y.conj(), X.conj()]])

        else:
            err_print = ('For `haar_mode="hermite"`, `beta` = 1, 2 or 4.',
                         'Given: {}'.format(beta))
            raise ValueError('\n'.join(err_print))

        _, U = la.eigh(A + A.conj().T)

    elif haar_mode == 'QR':

        if beta == 1:  # COE
            A = rng.randn(N, N)

        elif beta == 2:  # CUE
            A = rng.randn(N, N) + 1j * rng.randn(N, N)

        # elif beta==4:
        else:
            err_print = ('With `haar_mode="QR", `beta` = 1 or 2.',
                         'Given: {}'.format(beta))
            raise ValueError('\n'.join(err_print))

        # U, _ = la.qr(A)
        Q, R = la.qr(A)
        d = np.diagonal(R)
        U = np.multiply(Q, d / np.abs(d), Q)

    else:
        err_print = ('Invalid `haar_mode`.',
                     'Choose from `haar_mode="Hermite" or "QR".',
                     'Given: {}'.format(haar_mode))
        raise ValueError('\n'.join(err_print))

    return la.eigvals(U)


def mu_ref_unif_unit_circle_sampler_quindiag(beta=2, size=10,
                                             random_state=None):
    """
    .. see also::

        :cite:`KiNe04` Theorem 1
    """

    rng = check_random_state(random_state)

    if not (isinstance(beta, int) and (beta > 0)):
        raise ValueError('`beta` must be positive integer.\
                         Given: {}'.format(beta))

    alpha = np.zeros(size, dtype=np.complex_)

    # nu = 1 + beta*(N-1, N-2, ..., 0)
    for i, nu in enumerate(1 + beta * np.arange(size - 1, -1, step=-1)):
        gauss_vec = rng.randn(nu + 1)
        alpha[i] = (gauss_vec[0] + 1j * gauss_vec[1]) / la.norm(gauss_vec)

    rho = np.sqrt(1 - np.abs(alpha[:-1])**2)

    xi = np.zeros((size - 1, 2, 2), dtype=np.complex_)  # xi[0,..,N-1]
    xi[:, 0, 0], xi[:, 0, 1] = alpha[:-1].conj(), rho
    xi[:, 1, 0], xi[:, 1, 1] = rho, -alpha[:-1]

    # L = diag(xi_0, xi_2, ...)
    # M = diag(1, xi_1, x_3, ...)
    # xi[N-1] = alpha[N-1].conj()
    if size % 2 == 0:  # even
        L = sp.block_diag(xi[::2, :, :],
                          dtype=np.complex_)
        M = sp.block_diag([1.0, *xi[1::2, :, :], alpha[-1].conj()],
                          dtype=np.complex_)
    else:  # odd
        L = sp.block_diag([*xi[::2, :, :], alpha[-1].conj()],
                          dtype=np.complex_)
        M = sp.block_diag([1.0, *xi[1::2, :, :]],
                          dtype=np.complex_)

    return la.eigvals(L.dot(M).toarray())


###########
# Ginibre #
###########
def ginibre_sampler_full(N, random_state=None):
    """Compute the eigenvalues of a random complex standard Gaussian matrix"""

    rng = check_random_state(random_state)

    A = rng.randn(N, N) + 1j * rng.randn(N, N)

    return la.eigvals(A) / np.sqrt(2.0)



from string import ascii_lowercase
import numpy as np
from numpy.linalg import det, matrix_rank


def check_random_state(seed):
    """Turn seed into a np.random.RandomState instance
    Parameters
    ----------
    seed : None | int | instance of RandomState
        If seed is None, return the RandomState singleton used by np.random.
        If seed is an int, return a new RandomState instance seeded with seed.
        If seed is already a RandomState instance, return it.
        Otherwise raise ValueError.

    .. seealso::

        `Scikit learn source code <https://github.com/scikit-learn/scikit-learn/blob/7813f7efb/sklearn/utils/validation.py#L763>`_
    """
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, (int, np.integer)):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError('%r cannot be used to seed a numpy.random.RandomState'
                     ' instance' % seed)


def inner1d(arr1, arr2=None, axis=0):
    """ Efficient equivalent to ``(arr1**2).sum(axis)`` or ``(arr1*arr2).sum(axis)`` for ``arr1.shape == arr2.shape``.
    Expected to be used with arrays of same shape and mainly with 1D or 2D arrays but works for upto 26D arrays...

    If ``arr1.shape == arr2.shape``, then ``inner1d(arr1, arr2, arr1.ndim)`` replaces ``numpy.core.umath_tests.inner1d(arr1, arr2)``

    Examples:
    - To compute square norm of vector i.e. 1D array
    inner1d(arr) = np.einsum('i,i->', arr, arr)
                 = np.dot(arr, arr)
                 = (arr**2).sum()

    - To compute vector inner product i.e. 2 1D arrays
    inner1d(arr1, arr2) = np.einsum('i,i->', arr1, arr2)
                        = np.dot(arr1, arr2)
                        = (arr1*arr2).sum()

    - To compute square norm of cols/rows of 2D array
    inner1d(arr, axis=0/1)
        = np.einsum('ij,ij->j/i', arr, arr)
        = (arr**2).sum(axis=0/1)

    - To compute inner product between cols/rows of 2 arrays
    inner1d(arr1, arr2, axis=0/1)
        = np.einsum('ij,ij->j/i', arr1, arr2)
        = (arr1*arr2).sum(axis=0/1)
    """

    # if (arr2 is not None) and (arr1.shape != arr2.shape):
    #     raise ValueError('...with shapes {} {}'
    #                      .format(arr1.shape, arr2.shape))

    ndim = arr1.ndim
    sym = ascii_lowercase[:ndim]
    subscripts = sym + ',' + sym + '->' + sym.replace(sym[axis], '')

    if arr2 is None:
        return np.einsum(subscripts, arr1, arr1)
    else:
        return np.einsum(subscripts, arr1, arr2)


def det_ST(array, S, T=None):
    """ Compute :math:`\\det M_{S, T} = \\det [M_{ij}]_{i\\inS, j\\in T}`

    :param array:
        Matrix
    :type M:
        array_like

    :param S:
        collection of indices
    :type M:
        1D list, array_like

    :param T:
        collection of indices
    :type M:
        1D list, array_like, default None

    :return:
        - if `T is None` return :math:`\\det M_{S, S}`
        - else return :math:`\\det M_{S, T}`. If S=T=[], numpy convention = 1.0
    :rtype:
        float
    """

    if T is None:  # det M_SS = det M_S
        return det(array[np.ix_(S, S)])

    else:  # det M_ST, numpy deals with det M_[][] = 1.0
        return det(array[np.ix_(S, T)])


def is_square(array):

    if array is None:
        return None

    shape = array.shape
    if len(shape) == 2 and len(set(shape)) == 1:
        return array
    else:
        raise ValueError('array not 2D square: shape={}'.format(shape))


def is_symmetric(array):
    # Cheap test to check symmetry M^T = M

    if array is None:
        return None

    array = is_square(array)

    idx = np.arange(min(20, array.shape[0]))
    M = array[np.ix_(idx, idx)]
    if np.allclose(M.T, M):
        return array
    else:
        raise ValueError('array not symmetric: M.T != M')


def is_projection(array, col_idx=None):
    # Cheap test to check reproducing property: M^2 = M

    if array is None:
        return None

    array = is_square(array)

    if col_idx is None:
        col_idx = np.arange(min(20, array.shape[0]))

    M_j = array[:, col_idx]
    Mjj = array[col_idx, col_idx]

    if np.allclose(inner1d(M_j), Mjj):
        return array
    else:
        raise ValueError('array not seem to be a projection: M^2 != M')


def is_orthonormal_columns(array, col_idx=None):
    # Cheap test for checking orthonormality of columns of array: M.T M = I

    if array is None:
        return None

    if col_idx is None:
        col_idx = np.arange(np.min([5, array.shape[1]]))

    U = array[:, col_idx]

    if np.allclose(U.T.dot(U), np.eye(len(col_idx))):
        return array
    else:
        raise ValueError('array does not seem orthonormal: M.T M != I')


def is_equal_to_O_or_1(array, tol=1e-8):
    """Check if entries are **all** in :math:`\\{0, 1\\}`, for a given tolerance"""

    if array is None:
        return None

    equal_0 = np.abs(array) <= tol
    equal_1 = np.abs(1 - array) <= tol
    equal_0_or_1 = equal_0 ^ equal_1  # ^ = xor

    if np.all(equal_0_or_1):
        return array
    else:
        raise ValueError('array with entries not all in {0,1}')


def is_in_01(array, tol=1e-8):
    """Check if entries are **all** in :math:`[0, 1]`, for a given tolerance"""

    if array is None:
        return None
    elif np.all((-tol <= array) & (array <= 1.0 + tol)):
        return array
    else:
        raise ValueError('array with entries not all in [0,1]')


def is_geq_0(array, tol=1e-8):
    """Check if entries are **all** :math:`\\geq0`, for a given tolerance"""

    if array is None:
        return None
    elif np.all(array >= -tol):
        return array
    else:
        raise ValueError('array with entries not all >= 0')


def is_full_row_rank(array):
    # Check rank(M) = #rows

    if array is None:
        return None

    d, N = array.shape
    err_print = 'array (size = dxN) is not full row rank'

    if d > N:
        raise ValueError(err_print + 'd(={}) > N(={})'.format(d, N))
    else:
        rank = matrix_rank(array)
        if rank == d:
            return array
        else:
            raise ValueError(err_print + 'd(={}) != rank(={})'.format(d, rank))


def stable_invert_root(eigenvec, eigenval):
    """ Given eigendecomposition of a PSD matrix, compute a representation of the pseudo-inverse square root
    of the matrix using numerically stable operations. In particular, eigenvalues which are near-zero
    and the associated eigenvectors are dropped from the pseudo-inverse.
    """
    n = eigenvec.shape[0]

    if eigenvec.shape != (n, n) or eigenval.shape != (n,):
        raise ValueError('array sizes of {} eigenvectors and {} eigenvalues do not match'.format(eigenvec.shape, eigenval.shape))

    # threshold formula taken from pinv2's implementation of numpy/scipy
    thresh = eigenval.max() * max(eigenval.shape) * np.finfo(eigenval.dtype).eps
    stable_eig = np.logical_not(np.isclose(eigenval, 0., atol=thresh))
    m = sum(stable_eig)

    eigenvec_thin = eigenvec[:, stable_eig]
    eigenval_thin = eigenval[stable_eig]

    if eigenvec_thin.shape != (n, m) or eigenval_thin.shape != (m,):
        raise ValueError('array sizes of {} eigenvectors and {} eigenvalues do not match'.format(eigenvec.shape, eigenval.shape))

    eigenval_thin_inv_root = (1 / np.sqrt(eigenval_thin)).reshape(-1, 1)

    return eigenvec_thin, eigenval_thin_inv_root


def get_progress_bar(total=-1, disable=False, **kwargs):
    """Helper function to get a tqdm progress bar (or a simple fallback otherwise)"""
    class ProgBar(object):
        def __init__(self, total=-1, disable=False, **kwargs):
            self.disable = disable
            self.t = 0
            self.total = total
            self.debug_string = ""

        def __enter__(self):
            return self

        def __exit__(self, *args, **kwargs):
            pass

        def set_postfix(self, **kwargs):
            self.debug_string = ""
            for arg in kwargs:
                self.debug_string += "{}={} ".format(arg, kwargs[arg])

        def update(self):
            if not self.disable:
                self.t += 1
                print_str = "{}".format(self.t)

                if self.total > 0:
                    print_str += "/{}".format(self.total)

                print_str += ": {}".format(self.debug_string)

                if len(print_str) < 80:
                    print_str = print_str + " " * (80 - len(print_str))

                print(print_str, end='\r', flush=True)

            if self.t == self.total:
                print("")

    try:
        from tqdm import tqdm
        progress_bar = tqdm(total=total, disable=disable)
    except ImportError:
        progress_bar = ProgBar(total=total, disable=disable)

    return progress_bar


def evaluate_L_diagonal(eval_L, X):
    """Helper function to evaluate a likelihood function on a set of points (i.e. compute the diagonal of the L matrix)"""
    diag_eval = getattr(eval_L, "diag", None)
    if callable(diag_eval):
        return diag_eval(X)
    else:
        # inspired by sklearn.gaussian_process.kernels.PairwiseKernel
        return np.apply_along_axis(eval_L, 1, X).ravel()


def example_eval_L_linear(X, Y=None):
    X = np.atleast_2d(X)
    if Y is None:
        return X.dot(X.T)
    else:
        Y = np.atleast_2d(Y)
        return X.dot(Y.T)


def example_eval_L_polynomial(X, Y=None, p=2):
    if Y is None:
        ret = example_eval_L_linear(X)
        np.power(ret, p, out=ret)
        return ret
    else:
        ret = example_eval_L_linear(X, Y)
        np.power(ret, p, out=ret)
        return ret


def example_eval_L_min_kern(X, Y=None):

    X = np.atleast_2d(X)
    assert X.shape[1] == 1 and np.all((0 <= X) & (X <= 1))

    if Y is None:
        Y = X
    else:
        Y = np.atleast_2d(Y)
        assert Y.shape[1] == 1 and np.all((0 <= Y) & (Y <= 1))

    return np.minimum(np.repeat(X, Y.size, axis=1),
                      np.repeat(Y.T, X.size, axis=0))



# coding: utf8
""" Core functions for

- Uniform spanning trees

    * :func:`ust_sampler_wilson`
    * :func:`ust_sampler_aldous_broder`:

- Descent procresses :class:`Descent`:

    * :func:`uniform_permutation`

- :class:`PoissonizedPlancherel` measure

    * :func:`uniform_permutation`
    * :func:`RSK`: Robinson-Schensted-Knuth correspondande
    * :func:`xy_young_ru` young diagram -> russian convention coordinates
    * :func:`limit_shape`

.. seealso:

    `Documentation on ReadTheDocs <https://dppy.readthedocs.io/en/latest/exotic_dpps/index.html>`_
"""

import functools  # used for decorators to pass docstring

import numpy as np

from itertools import chain  # create graph edges from path

# For class PoissonizedPlancherel
from bisect import bisect_right  # for RSK

from dppy.utils import check_random_state


def ust_sampler_wilson(list_of_neighbors, root=None,
                       random_state=None):
    try:
        import networkx as nx
    except ImportError:
        raise ValueError('The networkx package is required to sample spanning trees (see setup.py).')

    rng = check_random_state(random_state)

    # Initialize the tree
    wilson_tree_graph = nx.Graph()
    nb_nodes = len(list_of_neighbors)

    # Initialize the root, if root not specified start from any node
    n0 = root if root else rng.choice(nb_nodes)  # size=1)[0]
    # -1 = not visited / 0 = in path / 1 = in tree
    state = -np.ones(nb_nodes, dtype=int)
    state[n0] = 1
    nb_nodes_in_tree = 1

    path, branches = [], []  # branches of tree, temporary path

    while nb_nodes_in_tree < nb_nodes:  # |Tree| = |V| - 1

        # visit a neighbor of n0 uniformly at random
        n1 = rng.choice(list_of_neighbors[n0])  # size=1)[0]

        if state[n1] == -1:  # not visited => continue the walk

            path.append(n1)  # add it to the path
            state[n1] = 0  # mark it as in the path
            n0 = n1  # continue the walk

        if state[n1] == 0:  # loop on the path => erase the loop

            knot = path.index(n1)  # find 1st appearence of n1 in the path
            nodes_loop = path[knot + 1:]  # identify nodes forming the loop
            del path[knot + 1:]  # erase the loop
            state[nodes_loop] = -1  # mark loopy nodes as not visited
            n0 = n1  # continue the walk

        elif state[n1] == 1:  # hits the tree => new branch

            if nb_nodes_in_tree == 1:
                branches.append([n1] + path)  # initial branch of the tree
            else:
                branches.append(path + [n1])  # path as a new branch

            state[path] = 1  # mark nodes in path as in the tree
            nb_nodes_in_tree += len(path)

            # Restart the walk from a random node among those not visited
            nodes_not_visited = np.where(state == -1)[0]
            if nodes_not_visited.size:
                n0 = rng.choice(nodes_not_visited)  # size=1)[0]
                path = [n0]

    tree_edges = list(chain.from_iterable(map(lambda x: zip(x[:-1], x[1:]),
                                              branches)))
    wilson_tree_graph.add_edges_from(tree_edges)

    return wilson_tree_graph


def ust_sampler_aldous_broder(list_of_neighbors, root=None,
                              random_state=None):

    try:
        import networkx as nx
    except ImportError:
        raise ValueError('The networkx package is required to sample spanning trees (see setup.py).')

    rng = check_random_state(random_state)

    # Initialize the tree
    aldous_tree_graph = nx.Graph()
    nb_nodes = len(list_of_neighbors)

    # Initialize the root, if root not specified start from any node
    n0 = root if root else rng.choice(nb_nodes)  # size=1)[0]
    visited = np.zeros(nb_nodes, dtype=bool)
    visited[n0] = True
    nb_nodes_in_tree = 1

    tree_edges = np.zeros((nb_nodes - 1, 2), dtype=np.int)

    while nb_nodes_in_tree < nb_nodes:

        # visit a neighbor of n0 uniformly at random
        n1 = rng.choice(list_of_neighbors[n0])  # size=1)[0]

        if visited[n1]:
            pass  # continue the walk
        else:  # create edge (n0, n1) and continue the walk
            tree_edges[nb_nodes_in_tree - 1] = [n0, n1]
            visited[n1] = True  # mark it as in the tree
            nb_nodes_in_tree += 1

        n0 = n1

    aldous_tree_graph.add_edges_from(tree_edges)

    return aldous_tree_graph


def uniform_permutation(N, random_state=None):
    """ Draw a perputation :math:`\\sigma \\in \\mathfrak{S}_N` uniformly at random using Fisher-Yates' algorithm

    .. seealso::

        - `Fisher–Yates_shuffle <https://en.wikipedia.org/wiki/Fisher%E2%80%93Yates_shuffle>_

        - `Numpy shuffle <https://github.com/numpy/numpy/blob/d429f0fe16c0407509b1f20d997bf94f1027f61b/numpy/random/mtrand.pyx#L4027>_`
    """
    rng = check_random_state(random_state)

    sigma = np.arange(N)
    for i in range(N - 1, 0, -1):  # reversed(range(1, N))
        j = rng.randint(0, i + 1)
        if j == i:
            continue
        sigma[j], sigma[i] = sigma[i], sigma[j]

    # for i in range(N - 1):
    #     j = rng.randint(i, N)
    #     sigma[j], sigma[i] = sigma[i], sigma[j]

    return sigma


def RSK(sequence):
    """Apply Robinson-Schensted-Knuth correspondence on a sequence of reals, e.g. a permutation, and return the corresponding insertion and recording tableaux.

    :param sequence:
        Sequence of real numbers
    :type sequence:
        array_like

    :return:
        :math:`P, Q` insertion and recording tableaux
    :rtype:
        list

    .. seealso::

        `RSK Wikipedia <https://en.wikipedia.org/wiki/Robinson%E2%80%93Schensted%E2%80%93Knuth_correspondence>`_
    """

    P, Q = [], []  # Insertion/Recording tableau

    for it, x in enumerate(sequence, start=1):

        # Iterate along the rows of the tableau P to find a place for the bouncing x and record the position where it is inserted
        for row_P, row_Q in zip(P, Q):

            # If x finds a place at the end of a row of P
            if x >= row_P[-1]:
                row_P.append(x)  # add the element at the end of the row of P
                row_Q.append(it)  # record its position in the row of Q
                break
            else:
                # find place for x in the row of P to keep the row ordered
                ind_insert = bisect_right(row_P, x)
                # Swap x with the value in place
                x, row_P[ind_insert] = row_P[ind_insert], x

        # If no room for x at the end of any row of P create a new row
        else:
            P.append([x])
            Q.append([it])

    return P, Q


def xy_young_ru(young_diag):
    """ Compute the xy coordinates of the boxes defining the young diagram, using the russian convention.

    :param young_diag:
        points
    :type  young_diag:
        array_like

    :return:
        :math:`\\omega(x)`
    :rtype:
        array_like
    """

    def intertwine(arr_1, arr_2):
        inter = np.empty((arr_1.size + arr_2.size,), dtype=arr_1.dtype)
        inter[0::2], inter[1::2] = arr_1, arr_2
        return inter

    # horizontal lines
    x_hor = intertwine(np.zeros_like(young_diag), young_diag)
    y_hor = np.repeat(np.arange(1, young_diag.size + 1), repeats=2)

    # vertical lines
    uniq, ind = np.unique(young_diag[::-1], return_index=True)
    gaps = np.ediff1d(uniq, to_begin=young_diag[-1])

    x_vert = np.repeat(np.arange(1, 1 + gaps.sum()), repeats=2)
    y_vert = np.repeat(young_diag.size - ind, repeats=gaps)
    y_vert = intertwine(np.zeros_like(y_vert), y_vert)

    xy_young_fr = np.column_stack(
        [np.hstack([x_hor, x_vert]), np.hstack([y_hor, y_vert])])

    rot_45_and_scale = np.array([[1.0, -1.0],
                                 [1.0, 1.0]])

    return xy_young_fr.dot(rot_45_and_scale.T)


def limit_shape(x):
    """ Evaluate :math:`\\omega(x)` the limit-shape function :cite:`Ker96`

    .. math::

        \\omega(x) =
        \\begin{cases}
            |x|, &\\text{if } |x|\\geq 2\\
            \\frac{2}{\\pi} \\left(x \\arcsin\\left(\\frac{x}{2}\\right) + \\sqrt{4-x^2} \\right) &\\text{otherwise } \\end{cases}

    :param x:
        points
    :type x:
        array_like

    :return:
        :math:`\\omega(x)`
    :rtype:
        array_like

    .. seealso::

        - :func:`plot_diagram <plot_diagram>`
        - :cite:`Ker96`
    """

    w_x = np.zeros_like(x)

    abs_x_gt2 = np.abs(x) >= 2.0

    w_x[abs_x_gt2] = np.abs(x[abs_x_gt2])
    w_x[~abs_x_gt2] = x[~abs_x_gt2] * np.arcsin(0.5 * x[~abs_x_gt2])\
                      + np.sqrt(4.0 - x[~abs_x_gt2]**2)
    w_x[~abs_x_gt2] *= 2.0 / np.pi

    return w_x



import numpy as np

from scipy.optimize import newton, brentq
from scipy.special import logsumexp, airy

from dppy.utils import check_random_state


def shift_pol(P, x0):
    """ Return polynomial P(x-x0) as poly1d object
    """

    X_x0 = np.poly1d([1, -x0])

    return sum(X_x0**n * c_n for n, c_n in enumerate(P.coeffs[::-1]))


def log_pdf_convex_quartic(x, P):
    """Evaluate :math:`\\ln` of

    .. math::

        \\pi(x) \\propto \\exp^{-(\\alpha x^4 + \\beta x^2 + \\gamma x)}
    """

    return -P(x)


def find_mode_convex_quartic(P):
    """ Compute mode of :math:`\\pi \\propto \\exp^{-P(x)}`,
    with :math:`P` convex polynomial.
    """

    roots = P.deriv(m=1).roots

    return roots[np.isreal(roots)].real[0]


def find_b_a_convex_quartic(P, mode):
    """ Compute :math:`b < 0 < a` solutions of :math:`\\pi(m+x)=\\frac{\\pi(m)}{4}`
    where :math:`\\pi \\propto \\exp^{-P(x)}` with :math:`P` convex polynomial.

    .. math::

        \\pi(m + x) = \\frac{\\pi(m)}{4}
        \\Longleftrightarrow
            P(m + x) - P(m) - 2\\ln(2) = 0
    """

    roots = (shift_pol(P, -mode) - P(mode) - 2.0 * np.log(2.0)).roots

    return np.sort(roots[np.isreal(roots)].real)


def log_pdf_convex_gen_gamma(x, shape, P):
    """Evaluate :math:`\\ln` of

    .. math::

        \\pi(x) \\propto x^{\\alpha-1} \\exp^{-(\\beta x^2 + \\gamma x)}
    """

    return (shape - 1) * np.log(x) - P(x)


def find_mode_convex_gen_gamma(shape, P):
    """ Compute mode of :math:`\\pi \\propto x^{a-1} \\exp^{-P(x)}`,
    with :math:`a>1` and :math:`P` convex polynomial.

    .. math::

        \\pi'(x) = 0
        \\Longleftrightarrow
            X P'(X) - (a-1) = 0
    """

    roots = (np.poly1d([1, 0]) * P.deriv(m=1) - (shape - 1)).roots
    real_roots = roots[np.isreal(roots)].real

    return real_roots[real_roots > 0][0]


def find_a_b_convex_gen_gamma(shape, P, mode):
    """Find solutions :math:`b<0, a>0` of :math:`\\pi(m+x)=\\frac{\\pi(m)}{4}`.

    For :math:`\\pi \\propto x^{\\alpha-1} \\exp^{-(\\beta x^2 + \\gamma x)}`,
     the equation is equivalent to

    .. math::

        (1-\\alpha) \\ln\\left(1+\\frac{x}{m}\\right)
            + \\beta x^2
            + (2\\beta m + \\gamma) x
            - 2\\ln(2)
        = 0
    """

    pol = shift_pol(P, -mode) - P(mode) - 2.0 * np.log(2.0)

    def f(x):
        return (shape - 1.0) * np.log(1.0 + x / mode) - pol(x)

    # Find b < 0 by bisection
    x0_b_left = - (1 - 1e-13) * mode
    x0_b_right = 0.0

    if np.sign(f(x0_b_left)) != np.sign(f(x0_b_right)):
        b = brentq(f, a=x0_b_left, b=x0_b_right)
    else:
        b = -0.5 * mode

    # Find a > 0 using Halley's method
    # initial point
    roots = pol.roots
    x0_a = np.max(roots[np.isreal(roots)].real)

    # 1st and 2nd derivatives
    d1_pol, d2_pol = pol.deriv(m=1), pol.deriv(m=2)

    def d1_f(x):
        return (shape - 1.0) / (mode + x) - d1_pol(x)

    def d2_f(x):
        return -(shape - 1.0) / (mode + x)**2 - d2_pol(x)

    a = newton(f, x0_a, fprime=d1_f, fprime2=d2_f)

    return np.array([b, a])  # b < 0 < a


def trunc_gaussian_sampler(b, mu, var, random_state=None):

    rng = check_random_state(random_state)

    it_max = int(1e7)

    for it in range(1, it_max + 1):
        X = rng.normal(loc=b, scale=np.sqrt(var))
        U = rng.rand()
        if (np.log(U) < -(b - mu) * (X - b) / var) and (X > b):
            break
    else:  # if no acceptation
        print('gen_gamma with alpha=1 not accepted\
                after {} iterations'.format(it))
        X = 1e-7

    return X, it


def gen_gamma_alpha_lt_1_sampler(shape, P, random_state=None):

    rng = check_random_state(random_state)

    it_max = int(1e2)
    for it in range(1, it_max + 1):

        X = rng.gamma(shape=shape, scale=1 / P[1], size=1)
        U = rng.rand()
        if np.log(U) < -P[2] * X**2:
            break

    else:
        print('gen_gamma with alpha<1 not accepted\
                after {} iterations'.format(it))
        pass
        # X = ...

    return X, it


def sampler_exact_convex_quartic(P, shape=None, random_state=None):
    """Sample from

    - shape is None

        .. math::

            \\pi(x)\\propto \\exp^{-(\\alpha x^4 + \\beta x^2 + \\gamma x)}

    - shape not None

        .. math::

            \\pi(x)\\propto x^{\\alpha-1} \\exp^{-(\\beta x**2 + \\gamma x)}

    .. seealso::

        :cite:`Dev12` Equation 2
        "A note on generating random variables with log-concave densities"
        `https://pdfs.semanticscholar.org/e9d3/f6e0862bc6893cdcce47fa6e8380d0fd765e.pdf <https://pdfs.semanticscholar.org/e9d3/f6e0862bc6893cdcce47fa6e8380d0fd765e.pdf>`_
    """

    # Compute mode of target density
    # Define f(x) = log_pi(x + mode) i.e. place the mode at the origin
    # Compute a>0, b<0 satisfying pi(m+x) >= pi(m)/4 >= pi(m+2x)

    rng = check_random_state(random_state)

    if shape is None:  # exp(-(alpha x^4 + beta x^2 + gamma x))
        mode = find_mode_convex_quartic(P)

        def log_f(x):
            return -P(x + mode)

        b, a = 0.5 * find_b_a_convex_quartic(P, mode)

    else:  # x^(shape-1) exp(-(beta x^2 + gamma x))

        if not P[2]:  # x^(shape-1) exp(- gamma x))
            return rng.gamma(shape, 1 / P[1]), 0

        else:
            if shape < 1.0:  # x^(shape-1) exp(-(beta x^2 + gamma x)): no mode
                return gen_gamma_alpha_lt_1_sampler(shape, P,
                                                    random_state=rng)

            elif shape == 1.0 and P[2]:
                # x^0 exp(-(beta x^2 + gamma x)) = truncated Gaussian [0, oo]
                return trunc_gaussian_sampler(b=0.0,
                                              mu=-0.5 * P[1] / P[2],
                                              var=0.5 / P[2],
                                              random_state=rng)

        # else there exists a mode, use [Dev12] domination
        mode = find_mode_convex_gen_gamma(shape, P)

        def log_f(x):
            return (shape - 1) * np.log(x + mode) - P(x + mode)

        b, a = 0.5 * find_a_b_convex_gen_gamma(shape, P, mode)

    _2b, _b, _a, _2a = 2 * b, b, a, 2 * a
    abs_b = abs(b)  # -b since b<0

    log_f_2b, log_f_b, log_f_0, log_f_a, log_f_2a =\
        log_f(np.array([_2b, _b, 0.0, _a, _2a]))
    log_f_b_2b, log_f_a_2a = log_f_b - log_f_2b, log_f_a - log_f_2a

    # Dominating function h(x) >= f(x)
    def log_h(x):
        if x < _2b:  # (2b-x)/b log_f(b) + (x-b)/b log_f(2b)
            return (_b - x) * (log_f_b_2b / _b) + log_f_b
        elif x < _b:  # 2b <= x < b
            return log_f_b
        elif x < _a:  # b <= x < a
            return log_f_0
        elif x < _2a:  # a <= x < 2a
            return log_f_a
        else:  # if x >= _2a, (2a-x)/a log_f(a) + (x-a)/a log_f(2a)
            return (_a - x) * (log_f_a_2a / _a) + log_f_a

    # Compute normalizing constant Z_h = int h
    # Z_h = int_[b, a] f(0)
    #     + int_[a,2a] f(a) + int_[2b,b] f(b)
    #     + int_[2a, +oo] exp.. + int_[-oo, 2b] exp..
    #
    # Z_h -= int_[-m, 2b] exp.. for gen_gamma

    a_lse = [log_f_0, log_f_a, log_f_b, log_f_2a, log_f_2b]
    b_lse = [_a + abs_b, _a, abs_b, _a / log_f_a_2a, abs_b / log_f_b_2b]

    if shape is not None:
        a_lse.append(log_f_b + (1 + mode / _b) * log_f_b_2b)
        b_lse.append(-abs_b / log_f_b_2b)
        trunc = mode + _2b
    else:
        trunc = np.inf

    log_Z_h = logsumexp(a=a_lse, b=b_lse)

    nb_iter_max = 100
    for iter_ in range(nb_iter_max):

        U, V, W = rng.rand(3)  # Unif_[0,1]
        log_V = np.log(V) + log_Z_h  # Incorporate Z_h in the LHS

        # Proposal mechanism:
        # 1. Choose a region C with proba 1/Z_h * \\int_C h (multinomial)
        # 2. Sample X ~ h restricted to C (uniform | exponential tails)

        if log_V <= logsumexp(a=a_lse[:1], b=b_lse[:1]):
            # propto int_[b, a] h = (a-b) f(0) = (a+|b|) f(0)
            # X ~ U([b, a])
            X = _b + (_a - _b) * U
        elif log_V <= logsumexp(a=a_lse[:2], b=b_lse[:2]):
            # propto int_[a, 2a] h = a f(a)
            # X ~ U([a, 2a])
            X = _a + _a * U
        elif log_V <= logsumexp(a=a_lse[:3], b=b_lse[:3]):
            # propto int_[2b, b] h = -b f(b) = |b| f(b)
            # X ~ U([2b, b])
            X = _b + _b * U
        elif log_V <= logsumexp(a=a_lse[:4], b=b_lse[:4]):
            # propto int_[2a, +oo] h = ...
            # X ~ 2a + Exp(log(f(a)/f(2a))/a)
            mu = _a / log_f_a_2a
            X = _2a - mu * np.log(U)
        else:
            # propto int_[-(oo or mode), 2b] h
            # X ~ 2b - Exp(log(f(b)/f(2b))/|b|)
            mu = abs_b / log_f_b_2b
            X = _2b + mu * np.log(1.0 - U * (1.0 - np.exp(-trunc / mu)))

        # Accept/reject X ~ h/Z_h with proba f(X)/h(X) = pi(X+m)/h(X)
        if np.log(W) <= log_f(X) - log_h(X):
            return X + mode, iter_

    else:
        print('Dev12 not accepted after {} iterations!'.format(iter_))
        return X + mode, iter_


def sampler_mala(x, V, sigma=0.01, nb_steps=20, random_state=None):
    """ Sample from :math:`\\pi(x) \\propto \\exp(-V(x))` with :math:`V` polynomial using Metropolis Adjusted Langevin Algorithm (MALA)
    """

    rng = check_random_state(random_state)

    x_ = x
    d1_V = V.deriv(m=1)

    for _ in range(nb_steps):

        Vx, grad_Vx = V(x_), d1_V(x_)
        y = x_ - 0.5 * sigma**2 * grad_Vx + sigma * rng.randn()
        Vy, grad_Vy = V(y), d1_V(y)

        acceptance = Vx - Vy\
                    + 0.5\
                        / sigma**2\
                        * ((x_ - y + 0.5 * sigma**2 * grad_Vy)**2
                            - (y - x_ + 0.5 * sigma**2 * grad_Vx)**2)

        if np.log(rng.rand()) < acceptance:
            x_ = y

    return x_


def polynomial_in_negative_log_conditional_a_coef(i, a, b, V):

    if len(V) > 7:
        str_ = ['Polynomial potentials V are allowed up to degree 6',
                'Given\n',
                ' '.join(['g_{}={}'.format(n, g_n)
                          for n, g_n in enumerate(V)])]
        raise ValueError(' '.join(str_))

    P = np.poly1d(0.0)
    i_2 = max(0, i - 2)

    if V[1]:
        P += V[1]\
            * np.poly1d(
                [1.0,  # X
                 0.0])  # 1

    if V[2]:
        P += V[2]\
            * np.poly1d(
                [1.0,  # X^2
                 0.0,  # X
                 0.0])  # 1

    if V[3]:
        P += V[3]\
            * np.poly1d(
                [1.0,  # X^3
                 0.0,  # X^2
                 3 * (b[i-1] + b[i]),  # X
                 0.0])  # 1

    if V[4]:
        P += V[4]\
            * np.poly1d(
                [1.0,  # X^4
                 0.0,  # X^3
                 4 * (b[i-1] + b[i]),  # X^2
                 4 * (a[i-1] * b[i-1] + a[i+1] * b[i]),  # X
                 0.0])  # 1

    if V[5]:
        P += V[5]\
            * np.poly1d(
                [1.0,  # X^5
                 0.0,  # X^4
                 5 * (b[i-1] + b[i]),  # X^3
                 5 * (a[i-1] * b[i-1] + a[i+1] * b[i]),  # X^2
                 5 * (b[i-1] * (a[i-1]**2 + b[i_2])
                        + (b[i-1] + b[i])**2
                        + b[i] * (a[i+1]**2 + b[i+1])),  # X
                 0.0])  # 1

    if V[6]:
        P += V[6]\
            * np.poly1d(
                [1.0,  # X^6
                 0.0,  # X^5
                 6 * (b[i-1] + b[i]),  # X^4
                 6 * (a[i-1] * b[i-1] + a[i+1] * b[i]),  # X^3
                 3 * (2 * b[i-1] * (a[i-1]**2 + b[i_2])
                        + 3 * (b[i-1] + b[i])**2
                        + 2 * b[i] * (a[i+1]**2 + b[i+1])),  # X^2
                 6 * (a[i_2] * b[i_2] * b[i-1]
                        + a[i-1] * b[i-1]
                                 * (a[i-1]**2 + 2*(b[i_2] + b[i-1] + b[i]))
                        + a[i+1] * b[i]
                                 * (a[i+1]**2 + 2*(b[i-1] + b[i] + b[i+1]))
                        + a[i+2] * b[i] * b[i+1]),  # X
                 0.0])  # 1

    return P


def polynomial_in_negative_log_conditional_b_coef(i, a, b, V):

    if len(V) > 7:
        str_ = ['Polynomial potentials V are allowed up to degree 6',
                'Given\n',
                ' '.join(['g_{}={}'.format(n, g_n)
                          for n, g_n in enumerate(V)])]
        raise ValueError(' '.join(str_))

    P = np.poly1d(0.0)

    if V[2]:
        P += V[2]\
            * np.poly1d(
                [2.0,  # X
                 0.0])  # 1

    if V[3]:
        P += V[3]\
            * np.poly1d(
                [3 * (a[i] + a[i+1]),  # X
                 0.0])  # 1

    if V[4]:
        P += V[4]\
            * np.poly1d(
                [2.0,  # X^2
                 4 * (a[i]**2 + a[i] * a[i+1] + a[i+1]**2
                        + b[i-1] + b[i+1]),  # X
                 0.0])  # 1

    if V[5]:
        P += V[5]\
            * np.poly1d(
                [5 * (a[i] + a[i+1]),  # X^2
                 5 * (a[i]**3 + a[i]**2 * a[i+1] + a[i] * a[i+1]**2 + a[i+1]**3
                        + b[i-1] * (a[i-1] + 2 * a[i] + a[i+1])
                        + b[i+1] * (a[i] + 2 * a[i+1] + a[i+2])),  # X
                 0.0])  # 1

    if V[6]:
        P += V[6]\
            * np.poly1d(
                [2.0,  # X^3
                 3 * (3 * a[i]**2 + 4 * a[i] * a[i+1] + 3 * a[i+1]**2
                        + 2 * (b[i-1] + b[i+1])),  # X^2
                 6 * (a[i]**4 + a[i]**3 * a[i+1] + a[i]**2 * a[i+1]**2
                        + a[i] * a[i+1]**3 + a[i+1]**4
                        + b[i-1] * (a[i-1]**2 + a[i-1] * a[i+1] + a[i+1]**2
                                    + a[i] * (2*a[i-1]+ 3*a[i] + 2*a[i+1])
                                    + b[i-2] + b[i-1] + b[i+1])
                        + b[i+1] * (a[i]**2 + a[i]*a[i+2] + a[i+2]**2
                                    + a[i+1] * (2*a[i]+ 3*a[i+1] + 2*a[i+2])
                                    + b[i+1] + b[i+2])),  # X
                 0.0])  # 1

    return P

"""Equilibrium densities"""


def equilibrium_x2_x4(g_2, g_4):
    """ The equilibrium measure associated to the :math:`\\beta`-ensemble with potential

    .. math::

        V(x) = V_t(x) = \\frac{g_2}{2} x^2 + \\frac{g_4}{4} x^4

    takes the form

    .. math::

        d \\mu(x) = \\frac{g_4}{2\\pi}(x^2 + b^2)\\sqrt{(a^2 - x^2)_+} \\ dx

    where :math:`a^2 = \\frac{2}{3g_4}\\left(\\sqrt{g_2^2+12g_4}-g_2\\right)` and :math:`b^2 = \\frac{1}{3g_4}\\left(\\sqrt{g_2^2+12g_4}+2g_2\\right)`

    .. seealso::

        - :cite:`DuKu06` p.2-3 `https://arxiv.org/pdf/math/0605201.pdf <https://arxiv.org/pdf/math/0605201.pdf>`_
        - :cite:`Mol` Example 3.3 `http://pcteserver.mi.infn.it/~molinari/RMT/RM2.pdf <http://pcteserver.mi.infn.it/~molinari/RMT/RM2.pdf>`_
    """

    if g_2 > -4 * np.sqrt(g_4 / 4):

        a2 = 2 / (3 * g_4) * (np.sqrt(g_2**2 + 12 * g_4) - g_2)
        b2 = 1 / (3 * g_4) * (np.sqrt(g_2**2 + 12 * g_4) + 2 * g_2)

        def dmu(x):
            return np.maximum(g_4 / (2 * np.pi)
                                * (x**2 + b2)
                                * np.sqrt(np.maximum(a2 - x**2, 0)),
                              0)

    else:
        b2, a2 = (-g_2 + 4 * np.sqrt(g_4 / 4) * np.array([-1, 1])) / g_4

        def dmu(x):
            return g_4 / (2 * np.pi)\
                       * np.abs(x)\
                       * np.sqrt(np.maximum((a2 - x**2)*(x**2 - b2), 0))

    support = np.sqrt(a2) * np.array([-1, 1])

    return dmu, support


def cdf_equilibrium_x2_x4(g_2, g_4):
    """ Cumulative distribution function of the equilibrium distribution associated to the :math:`\\beta`-ensemble with potential

    .. math::

        V(x) = V_t(x) = \\frac{g_2}{2} x^2 + \\frac{g_4}{4} x^4
    """

    a2 = 2 / (3 * g_4) * (np.sqrt(g_2**2 + 12 * g_4) - g_2)
    a = np.sqrt(a2)

    b2 = 1 / (3 * g_4) * (np.sqrt(g_2**2 + 12 * g_4) + 2 * g_2)

    def cdf(x):

        x[x < -a] = -a
        x[x > a] = a
        X = np.arcsin(x / a)
        # return g_4 * a2 / (8 * np.pi) * ((X + np.pi) * (a2 + 4 * b2) / 4
        #                                  - a2 * np.sin(4 * X) / 8
        #                                  + b2 * np.sin(2 * X))

        return g_4 * a2 / (8 * np.pi)\
                * ((a2 + 4 * b2) / 2 * (X + np.pi / 2)
                   + b2 * np.sin(2 * X)
                   - a2 / 8 * np.sin(4 * X))

    support = a * np.array([-1, 1])

    return cdf, support


def equilibrium_x2m(m, g_2m):
    """ The equilibrium measure associated to the :math:`\\beta`-ensemble with potential

    .. math::

        V(x) = \\underbrace{\\frac{g_{2m}}{2m}}_{=t} x^{2m}

    takes the form

    .. math::

        d \\mu(x) = \\frac{mt}{\\pi} \\sqrt{(a^2 - x^2)_+} h_1(x) \\ dx

    where
    :math:`a^2 = \\left( m t \\prod_{l=1}^{m} \\frac{2l-1}{2l} \\right)^{-\\frac{1}{m}}`
    and
    :math:`h_1(x) = x^{2(m-1)} + \\sum_{j=1}^{m-1} x^{2(m-j-1)} a^{2j} \\prod_{l=1}^{j} \\frac{2l - 1}{2l}`

    .. seealso::

        - :cite:`Dei00` Proposition 6.156 "Orthogonal polynomials and random matrices : a Riemann-Hilbert approach", ISBN 0821826956.
    """

    t = g_2m / (2 * m)
    c_prod = np.cumprod([(2 * l - 1) / (2 * l) for l in range(1, m + 1)])
    a2 = 1 / (m * t * c_prod[-1])**(1 / m)

    h1 = np.poly1d(0.0)
    h1[2 * (m - 1)] = 1.0

    if m > 1:
        h1._coeffs[2::2] = a2**np.arange(1, m) * c_prod[:-1]

    def mu_eq(x):
        return np.maximum(m * t / np.pi
                            * np.sqrt(np.maximum(a2 - x**2, 0))
                            * h1(x),
                          0)

    support = np.sqrt(a2) * np.array([-1, 1])

    return mu_eq, support


def cdf_equilibrium_x2m(m, g_2m):
    """ Cumulative distribution function of the equilibrium distribution associated to the :math:`\\beta`-ensemble with potential

    .. math::

        V(x) = \\underbrace{\\frac{g_{2m}}{2m}}_{=t} x^{2m}

    """

    t = g_2m / (2 * m)
    c_prod = np.cumprod([(2 * l - 1) / (2 * l) for l in range(1, m + 1)])
    a2 = 1 / (m * t * c_prod[-1])**(1 / m)

    a = np.sqrt(a2)

    if m == 2:
        def cdf(x):

            x[x < -a] = -a
            x[x > a] = a
            X = np.arcsin(x / a)
            return m * t * a**4 / (8 * np.pi) \
                   * (3 * (X + np.pi / 2)
                      + np.sin(2 * X)
                      - 1 / 4 * np.sin(4 * X))

    elif m == 3:
        def cdf(x):

            x[x < -a] = -a
            x[x > a] = a
            X = np.arcsin(x / a)
            return m * t * a**6 / (16 * np.pi) \
                   * (5 * (X + np.pi / 2)
                      + 5 / 4 * np.sin(2 * X)
                      - 1 / 2 * np.sin(4 * X)
                      + 1 / 12 * np.sin(6 * X))

    else:
        raise NotImplementedError('m != 2, 3')

    support = a * np.array([-1, 1])

    return cdf, support


def equilibrium_ClItKr10():
    """.. math::

        V(x) = \\frac{1}{20} x^4 - \\frac{4}{15}x^3 + \\frac{1}{5}x^2 + \\frac{8}{5}x

    .. math::

        d \\mu(x) = \\frac{1}{10\\pi} \\sqrt{[(x+2)(2-x)^5]_+} \\ dx

    .. seealso::

        - :cite:`ClItKr10` Example 1.2 HIGHER-ORDER ANALOGUES OF THE TRACY-WIDOM DISTRIBUTION
        - :cite:`OlNaTr14` Section 3.2 https://arxiv.org/pdf/1404.0071.pdf
    """

    def mu_eq(x):
        return 1 / (10 * np.pi) * np.sqrt(np.maximum((x + 2) * (2 - x)**5, 0))

    support = 2.0 * np.array([-1, 1])

    return mu_eq, support


class TracyWidom(object):
    """ Implements computation of the cumulative distribution function (CDF) of the Tracy Widom 2 distribution based on the work of Bornemann [2010] https://arxiv.org/pdf/0804.2543.pdf

    .. seealso::

        :ref:`another implementation <https://gist.github.com/yymao/7282002>`
    """
    def __init__(self):
        self.quad_order = None
        self.x_quad, self.w_quad = None, None

    def change_of_variable(self, s):  # , old_implem=False):
        """Bornemann Equation 7.5"""

        def phi(x):
            # if old_implem:
            #     return s + 10 * np.tan(0.5 * np.pi * x)
            # else:
            return np.atleast_1d(s)[:, None] + 10 * np.tan(0.5 * np.pi * x)

        def d_phi(x):
            return 10 * np.pi / 2 * (1 + np.tan(0.5 * np.pi * x)**2)

        return phi, d_phi

    def kernel(self, s):  # , old_implem=False):
        """ operator is restricted to L^2(0,1), so you need to change variables first
        """
        phi, d_phi = self.change_of_variable(s)

        # if old_implem:

        #     def K(x, y):
        #         K_Ai_xy = self.airy_kernel(phi(x), phi(y))

        #         return np.sqrt(d_phi(x) * d_phi(y)) * K_Ai_xy

        # else:

        def K(X, Y=None, eps_xy=1e-5):
            # Compute K[x, y]_{x, y \in X}
            # K_xy = (A_x * dA_y - dA_x * A_y) / (x - y)
            X = np.ravel(X)
            size_X = len(X)

            x = phi(X)
            sqrt_dphi_x = np.sqrt(d_phi(X))
            A_x, dA_x, _, _ = airy(x)
            # y = x
            A_y, dA_y = A_x, dA_x

            outer = 'ki, kj -> kij'
            K_xy = np.einsum(outer, A_x, dA_y) - np.einsum(outer, dA_x, A_y)

            # divide upper triangular part and symmetrize
            # K_xy /= x_y, where x_y = [x_i - y_j]_ij (recall y = x)
            x_y = x[0][:, None] - x[0]
            idx = np.triu_indices_from(x_y, k=1)

            K_xy[:, idx[0], idx[1]] /= x_y[idx[0], idx[1]]  # K_xy /= x-y
            K_xy[:, idx[1], idx[0]] = K_xy[:, idx[0], idx[1]]  # K_yx = K_xy

            # fill the diagonal K_xx
            # set y = x - eps
            # K_xx = (A_x * dA_x-eps - dA_x * A_x-eps) / eps
            A_y, dA_y, _, _ = airy(x - eps_xy)
            diag = range(size_X)
            K_xy[:, diag, diag] = (A_x * dA_y - dA_x * A_y) / eps_xy

            return sqrt_dphi_x[:, None] * K_xy * sqrt_dphi_x

        return K

    def cdf(self, x, quad_order=50):  # , old_implem=False):
        """Compute cumulative distribution function :math:`F_{TW_2}` of the Tracy-Widom :math:`TW_2` distribution

        :param x: points where
        :type x: float or 1D array-like

        :param quad_order: order (degree) of the quadrature used to compute the underlying Fredholm determinant.
        :type quad_order: int
        """

        if self.quad_order != quad_order:
            self.quad_order = quad_order
            self.x_quad, self.w_quad = self.compute_quadrature(quad_order)

        return self.fredholm_determinant(self.kernel(x),
                                         self.x_quad,
                                         self.w_quad)  # , old_implem)

    @staticmethod
    def airy_kernel(x, y, eps_xy=1e-5):
        """evaluate Airy kernel pointwise
        https://en.wikipedia.org/wiki/Tracy%E2%80%93Widom_distribution
        """

        # hack to extend the kernel by continuity on the diagonal
        if x == y:
            y += eps_xy

        Ai_x, dAi_x, _, _ = airy(x)
        Ai_y, dAi_y, _, _ = airy(y)

        return (Ai_x * dAi_y - Ai_y * dAi_x) / (x - y)

    @staticmethod
    def compute_quadrature(order, type='Legendre'):

        x, w = np.polynomial.legendre.leggauss(order)
        return 0.5 * (x + 1), 0.5 * w

    @staticmethod
    def fredholm_determinant(kernel, x_quad, w_quad):  # , old_implem):
        """ This implements the numerical evaluation of Fredholm determinants as in [Bornemann, 2010] Equation 6.1

        .. math::

            \\det (I - K)_L^2[0,1] \\approx \\det (I - [K(x, y)]_xy)
        """

        # if old_implem:
        #     K = np.array([[kernel(x, y) for x in x_quad] for y in x_quad])
        # else:
        K = kernel(x_quad, x_quad)
        shp = K.shape

        d = np.sqrt(w_quad)

        Id = np.broadcast_to(np.eye(*shp[1:]), shp)

        return np.linalg.det(Id - d[:, None] * K * d)
